#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Assessment results report tasks module """
import json
from celery.exceptions import Reject

from django.core.files.base import ContentFile
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Avg
from django.db.models import Max
from django.db.models import Min
from django.db.models import Count

from tesla_ce import celery_app
from tesla_ce import models


@celery_app.task(ignore_result=True, bind=True)
def update_learner_activity_session_report(self, learner_id, activity_id, session_id, force_update=False):
    """
        Create a report for an assessment session
        :param self: Task object
        :param learner_id: Learner id
        :param activity_id: Activity id
        :param session_id: Assessment session id
        :param force_update: If True, compute data even if there are pending requests, otherwise is skipped
    """
    # Get the report
    try:
        report = models.ReportActivity.objects.get(
            activity_id=activity_id,
            learner_id=learner_id
        )
    except models.ReportActivity.DoesNotExist:
        report = models.ReportActivity.objects.create(
            activity_id=activity_id,
            learner_id=learner_id
        )

    # Get the session report
    try:
        session_report = models.ReportActivitySession.objects.get(
            session_id=session_id
        )
    except models.ReportActivitySession.DoesNotExist:
        session_report = models.ReportActivitySession.objects.create(
            session_id=session_id,
            report=report,
            integrity_level=0,
            identity_level=0,
            content_level=0
        )

    # Get session requests statistics
    session_report.total_requests = models.Request.objects.filter(session_id=session_id).count()
    session_report.valid_requests = models.Request.objects.filter(session_id=session_id, status=3).count()
    session_report.pending_requests = models.Request.objects.filter(session_id=session_id, status__lt=3).count()
    session_report.processed_requests = models.Request.objects.filter(session_id=session_id, status__gte=3).count()

    # If all the data for this session is processed, compute session detail
    if session_report.pending_requests == 0 or force_update:
        results_qs = models.RequestResult.objects.filter(request__session_id=session_id,
                                                         request__learner_id=learner_id)
        data = {}
        results_stats = results_qs.filter(status=1).values('instrument').annotate(Max('code'),
                                                                                  Avg('result'),
                                                                                  Count('instrument'))
        for value in results_stats:
            if value['instrument'] not in data:
                data[value['instrument']] = {
                    'total': 0,
                    'valid': 0,
                    'confidence': None,
                    'result': None,
                    'code': None,
                }
            if value['code__max'] > 0:
                data[value['instrument']]['code'] = value['code__max'] + 1
            else:
                data[value['instrument']]['code'] = value['code__max']
            data[value['instrument']]['result'] = value['result__avg']
        num_requests = results_qs.values('instrument').annotate(Count('instrument'))
        for value in num_requests:
            if value['instrument'] not in data:
                data[value['instrument']] = {
                    'total': 0,
                    'valid': 0,
                    'confidence': None,
                    'result': None,
                    'code': None,
                }
            data[value['instrument']]['total'] = value['instrument__count']

        valid_requests = results_qs.filter(status=1).values('instrument').annotate(Count('instrument'))
        for value in valid_requests:
            if value['instrument'] not in data:
                data[value['instrument']] = {
                    'total': 0,
                    'valid': 0,
                    'confidence': None,
                    'result': None,
                    'code': None,
                }
            data[value['instrument']]['valid'] = value['instrument__count']
        for inst_id in data:
            if data[inst_id]['total'] > 0:
                data[inst_id]['confidence'] = data[inst_id]['valid'] / data[inst_id]['total']
            data[inst_id]['requests'] = json.dumps(
                list(results_qs.values_list('created_at', 'result').order_by('created_at')),
                cls=DjangoJSONEncoder)

        session_data = {
            'instruments': list(results_qs.values_list('instrument', flat=True).distinct()),
            'alerts': json.dumps(
                list(models.Alert.objects.filter(session_id=session_id).order_by('created_at')),
                cls=DjangoJSONEncoder),
            'data': data
        }
        data_path = '{}/results/{}/{}/{}/session_{}.json'.format(
                session_report.session.activity.vle.institution_id,
                session_report.session.learner.learner_id,
                session_report.session.activity.course.id,
                session_report.session.activity.id,
                session_id
            )
        session_report.data.save(data_path,
                                 ContentFile(json.dumps(session_data, cls=DjangoJSONEncoder).encode('utf-8')))

    session_report.save()


@celery_app.task(ignore_result=True, bind=True)
def update_learner_activity_instrument_report(self, learner_id, activity_id, instrument_id):

    # Get the report
    try:
        report = models.ReportActivity.objects.get(
            activity_id=activity_id,
            learner_id=learner_id
        )
    except models.ReportActivity.DoesNotExist:
        report = models.ReportActivity.objects.create(
            activity_id=activity_id,
            learner_id=learner_id
        )
    # Get the instrument report detail
    try:
        instrument_report = models.ReportActivityInstrument.objects.get(
            report=report,
            instrument_id=instrument_id
        )
    except models.ReportActivityInstrument.DoesNotExist:
        try:
            instrument_report = models.ReportActivityInstrument.objects.create(
                report=report,
                instrument_id=instrument_id,
                enrolment=0,
                confidence=0,
                result=0
            )
        except Exception:
            # In the unprovable case two tasks try to access report the first time, one of them will fail on creation
            self.retry(countdown=(self.request.retries + 1) * 15, max_retries=5)
            return
    results_qs = models.RequestResult.objects.filter(request__activity_id=activity_id,
                                                     instrument_id=instrument_id,
                                                     request__learner_id=learner_id)

    # Compute the final alert levels from result codes
    code_result = results_qs.filter(status=1).aggregate(Max('code'), Avg('result'))
    instrument_report.result = round(code_result['result__avg'] * 100)
    instrument_report.confidence = round(results_qs.filter(status=1).count() / results_qs.count() * 100)
    code = code_result['code__max']
    if code > 0:
        # Unless the code is 0 (PENDING), move to the alerts scale
        code += 1
    # Update levels where this instrument applies
    if instrument_report.instrument.identity:
        instrument_report.identity_level = code
    else:
        instrument_report.identity_level = 1  # No data
    if instrument_report.instrument.originality or instrument_report.instrument.authorship:
        instrument_report.content_level = code
        instrument_report.authorship_level = code
    else:
        instrument_report.content_level = 1  # No data
        instrument_report.authorship_level = 1  # No data
    if instrument_report.instrument.integrity:
        instrument_report.integrity_level = code
    else:
        instrument_report.integrity_level = 1  # No data
    # Get mean enrolment value from used providers
    instrument_report.enrolment = 0
    if instrument_report.instrument.requires_enrolment:
        instrument_report.enrolment = round(models.Enrolment.objects.filter(
            learner_id=learner_id,
            provider_id__in=results_qs.values_list(
                'request__requestproviderresult__provider_id',
                flat=True).distinct()
        ).aggregate(Min('percentage'))['percentage__min'] * 100)
    instrument_report.save()

    # Update the audit data
    instrument_report.update_audit()

    # Update the report
    update_learner_activity_report.apply_async((learner_id, activity_id,))


@celery_app.task(ignore_result=True)
def update_learner_activity_report(learner_id, activity_id, force_update=False):
    """
        Update the report for an activity and learner
        :param learner_id: Learner id
        :param activity_id: Activity id
        :param force_update: If True, compute data even if there are pending requests, otherwise is skipped
    """
    # Skip update if no new data is present
    try:
        report = models.ReportActivity.objects.get(
            activity_id=activity_id,
            learner_id=learner_id
        )
    except models.ReportActivity.DoesNotExist:
        # This task is executed always after instrument results tasks. It is not possible that report does not exist
        raise Reject('Report does not exists')

    # If updated is forced, build instrument results
    if force_update:
        for inst_report in report.reportactivityinstrument_set.all():
            update_learner_activity_instrument_report(learner_id, activity_id, inst_report.instrument.id)

    # Check if there are new results
    last_update = report.reportactivityinstrument_set.aggregate(Max('updated_at'))['updated_at__max']
    if last_update is not None and report.updated_at > last_update and not force_update:
        raise Reject('No new results')

    stats = report.reportactivityinstrument_set.aggregate(Max('identity_level'),
                                                          Max('content_level'),
                                                          Max('integrity_level'))
    # Update levels
    report.identity_level = stats['identity_level__max']
    report.content_level = stats['content_level__max']
    report.integrity_level = stats['integrity_level__max']

    # Create the report data
    report_data = {
        'sessions': [],
        'documents': []
    }

    # Update sessions' results
    sessions = models.AssessmentSession.objects.filter(
        activity_id=activity_id,
        learner_id=learner_id
    ).values_list('id', flat=True)
    for session_id in list(sessions):
        update_learner_activity_session_report(learner_id=learner_id, activity_id=activity_id, session_id=session_id)

    for session in report.reportactivitysession_set.all().order_by('created_at'):
        closed_at = None
        if session.closed_at is not None:
            closed_at = session.closed_at.isoformat()
        session_data = None
        try:
            if session.data is not None:
                session_data = json.loads(session.data.read().decode())
        except Exception:
            # If there is any error reading the session data, just left it as None
            pass
        report_data['sessions'].append({
            'id': session.id,
            'pending_requests': session.pending_requests,
            'valid_requests': session.valid_requests,
            'processed_requests': session.processed_requests,
            'total_requests': session.total_requests,
            'created_at': session.created_at.isoformat(),
            'closed_at': closed_at,
            'identity_level': session.identity_level,
            'integrity_level': session.integrity_level,
            'content_level': session.content_level,
            'data': session_data
        })

    # Add results for attachments without session
    documents = models.Request.objects.filter(learner_id=learner_id,
                                              activity_id=activity_id,
                                              session__isnull=True).all()
    for doc in documents:
        doc_data = {}
        instruments = []
        for doc_res in doc.requestresult_set.values_list('instrument_id', 'status', 'result'):
            instruments.append(doc_res[0])
            doc_data[doc_res[0]] = {
                'status': doc_res[1],
                'result': doc_res[2]
            }
        report_data['documents'].append({
            'instruments': instruments,
            'results': doc_data,
            'created_at': doc.created_at.isoformat()
        })

    # Store report data
    data_path = '{}/results/{}/{}/{}/report.json'.format(
        report.activity.vle.institution_id,
        report.learner.learner_id,
        report.activity.course.id,
        report.activity.id
    )
    report.data.save(data_path,
                     ContentFile(json.dumps(report_data, cls=DjangoJSONEncoder).encode('utf-8')))
    report.save()


@celery_app.task(ignore_result=True)
def update_activity_report(activity_id):
    pass


@celery_app.task(ignore_result=True)
def update_course_report(course_id):
    pass



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

""" VLe nested serializers package """
from .activity import VLECourseActivitySerializer
from .activity_instrument import VLECourseActivityInstrumentSerializer
from .activity_report import VLECourseActivityReportExtendedSerializer
from .activity_report import VLECourseActivityReportSerializer
from .assessment_session import VLENewAssessmentSessionBodySerializer
from .assessment_session import VLENewAssessmentSessionSerializer
from .attachment import VLECourseActivityAttachmentSerializer
from .course import VLECourseSerializer
from .instructor import VLECourseInstructorSerializer
from .instruments import VLEInstrumentSerializer
from .launcher import VLELauncherBodySerializer
from .launcher import VLELauncherDataSerializer
from .learner import VLECourseActivityLearnerSerializer
from .learner import VLECourseLearnerSerializer
from .request import VLECourseActivityLearnerRequestSerializer
from .vle import VLESerializer

__all__ = [
    "VLESerializer",
    "VLECourseSerializer",
    "VLECourseActivitySerializer",
    "VLECourseActivityInstrumentSerializer",
    "VLECourseActivityAttachmentSerializer",
    "VLECourseLearnerSerializer",
    "VLECourseActivityLearnerSerializer",
    "VLECourseActivityLearnerRequestSerializer",
    "VLECourseInstructorSerializer",
    "VLENewAssessmentSessionSerializer",
    "VLENewAssessmentSessionBodySerializer",
    "VLECourseActivityReportSerializer",
    "VLECourseActivityReportExtendedSerializer",
    "VLEInstrumentSerializer",
    "VLELauncherBodySerializer",
    "VLELauncherDataSerializer",
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA Identity Provider Serializers """
from rest_framework import serializers


class LearnerIdSerializer(serializers.Serializer):
    """ Body serializer for a single mail."""

    mail = serializers.EmailField(required=True, allow_blank=False, allow_null=False)

    class Meta:
        fields = "mail"


class MultipleLearnerIdSerializer(serializers.Serializer):
    """ Body serializer for a single mail."""

    mails = serializers.ListField(required=True, child=serializers.EmailField(), allow_null=False)

    class Meta:
        fields = "mails"


class LearnerSerializer(serializers.Serializer):
    """ Learner serialization """

    tesla_id = serializers.UUIDField(source='learner_id', read_only=True)
    mail = serializers.EmailField(source='email', read_only=True)

    class Meta:
        fields = ["teala_id", "mail"]



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Institution Administrator actions for Use Case tests """
from django.utils import timezone

from tests import auth_utils
from tests.conftest import get_random_string


def api_create_institution_legal_admin(inst_admin):
    """
        An institution administrator creates a new legal admin
        :param inst_admin: Institution admin object
        :return: New legal administrator object
    """
    # Authenticate with admin credentials
    client = auth_utils.client_with_user_credentials(inst_admin['email'], inst_admin['password'])

    # Get the user profile
    profile = auth_utils.get_profile(client)

    # Create a new user
    user_name = get_random_string(10)
    password = get_random_string(10)
    institution_id = profile['institution']['id']
    email = '{}@tesla-ce.eu'.format(user_name)
    user_data = {
        'username': user_name,
        'uid': user_name,
        'email': email,
        'first_name': user_name[:5],
        'last_name': user_name[5:],
        'institution': institution_id,
        'login_allowed': True,
        'password': password,
        'password2': password,
        'inst_admin': False,
        'legal_admin': True
    }
    user_create_resp = client.post('/api/v2/institution/{}/user/'.format(institution_id),
                                   data=user_data, format='json')
    assert user_create_resp.status_code == 201

    # Get institution legal admin object
    legal_admin_user = user_create_resp.data
    assert legal_admin_user['legal_admin']

    # Return credentials
    return {
        'email': email,
        'password': password
    }


def api_create_institution_send_admin(inst_admin):
    """
        An institution administrator creates a new SEND admin
        :param inst_admin: Institution admin object
        :return: New SEND administrator object
    """
    # Authenticate with admin credentials
    client = auth_utils.client_with_user_credentials(inst_admin['email'], inst_admin['password'])

    # Get the user profile
    profile = auth_utils.get_profile(client)

    # Create a new user
    user_name = get_random_string(10)
    password = get_random_string(10)
    institution_id = profile['institution']['id']
    email = '{}@tesla-ce.eu'.format(user_name)
    user_data = {
        'username': user_name,
        'uid': user_name,
        'email': email,
        'first_name': user_name[:5],
        'last_name': user_name[5:],
        'institution': institution_id,
        'login_allowed': True,
        'password': password,
        'password2': password,
        'inst_admin': False,
        'send_admin': True
    }
    user_create_resp = client.post('/api/v2/institution/{}/user/'.format(institution_id),
                                   data=user_data, format='json')
    assert user_create_resp.status_code == 201

    # Get institution SEND admin object
    send_admin_user = user_create_resp.data
    assert send_admin_user['send_admin']

    # Return credentials
    return {
        'email': email,
        'password': password
    }


def api_create_ic(admin, version):
    """
        A legal administrator of the institution creates the Informed Consent using the API
        :param admin: Credentials for a user with legal administration rights
        :param version: Version of the Informed Consent
    """
    # Authenticate with admin credentials
    client = auth_utils.client_with_user_credentials(admin['email'], admin['password'])

    # Get the user profile
    profile = auth_utils.get_profile(client)

    # Create an IC
    institution_id = profile['institution']['id']
    ic_data = {
        'version': version,
        'valid_from': timezone.now() - timezone.timedelta(days=1),
    }
    ic_create_resp = client.post('/api/v2/institution/{}/ic/'.format(institution_id),
                                 data=ic_data, format='json')
    assert ic_create_resp.status_code == 201
    new_ic = ic_create_resp.data
    assert new_ic['institution']['id'] == institution_id

    # Create a new IC document
    ic_doc_data = {
        'language': 'en',
        'html': '<h1>Informed Consent</h1>',
    }
    # TODO: Add PDF document
    version_id = new_ic['id']
    ic_doc_create_resp = client.post(
        '/api/v2/institution/{}/ic/{}/document/'.format(institution_id, version_id),
        data=ic_doc_data, format='json'
    )
    assert ic_doc_create_resp.status_code == 201
    new_ic_doc = ic_doc_create_resp.data

    assert new_ic_doc['consent']['id'] == version_id
    assert new_ic_doc['language'] == 'en'


def api_create_send_categories(admin, disabled_inst=None):
    """
        A SEND administrator of the institution defines the SEND categories using the API
        :param admin: Credentials for a user with SEND administration rights
        :param disabled_inst: List of instruments to disable
    """
    if disabled_inst is None:
        disabled_inst = []

    # Authenticate with admin credentials
    client = auth_utils.client_with_user_credentials(admin['email'], admin['password'])

    # Get the user profile
    profile = auth_utils.get_profile(client)

    # Create an SEND category disabling KS
    institution_id = profile['institution']['id']
    send_cat_resp = client.post(
        '/api/v2/institution/{}/send/'.format(institution_id),
        data={
            'description': 'Test SEND category disabling KS',
            'data': {
                'enabled_options': [],
                'disabled_instruments': disabled_inst
            }
        },
        format='json'
    )
    assert send_cat_resp.status_code == 201
    send_category = send_cat_resp.data

    return send_category


def api_enable_direct_registration_vle(inst_admin):
    """
        Institution enables direct learners and instructors registration by VLE
        :param inst_admin: Institution admin credentials
        :return: VLE object
    """
    # Authenticate with admin credentials
    client = auth_utils.client_with_user_credentials(inst_admin['email'], inst_admin['password'])

    # Get the user profile
    profile = auth_utils.get_profile(client)

    # Enable direct registration
    institution_id = profile['institution']['id']
    inst_mod_resp = client.patch(
        '/api/v2/institution/{}/'.format(institution_id),
        data={
            'disable_vle_user_creation': False,
            'disable_vle_learner_creation': False,
            'disable_vle_instructor_creation': False,
        },
        format='json'
    )
    assert inst_mod_resp.status_code == 200
    assert not inst_mod_resp.data['disable_vle_user_creation']
    assert not inst_mod_resp.data['disable_vle_learner_creation']
    assert not inst_mod_resp.data['disable_vle_instructor_creation']


def api_register_vle(inst_admin):
    """
        Institution administrator register a new VLE
        :param inst_admin: Institution admin credentials
        :return: VLE object
    """
    # Authenticate with admin credentials
    client = auth_utils.client_with_user_credentials(inst_admin['email'], inst_admin['password'])

    # Get the user profile
    profile = auth_utils.get_profile(client)

    # Register a VLE
    vle_name = get_random_string(10)
    vle_data = {
        'type': 0,  # 0 is the code for Moodle
        'name': vle_name,
        'url': '{}.tesla-ce-eu'.format(vle_name),
        'client_id': get_random_string(15)
    }
    institution_id = profile['institution']['id']
    vle_create_resp = client.post(
        '/api/v2/institution/{}/vle/'.format(institution_id),
        data=vle_data, format='json'
    )
    assert vle_create_resp.status_code == 201
    vle = vle_create_resp.data

    assert 'credentials' in vle
    assert 'role_id' in vle['credentials']
    assert 'secret_id' in vle['credentials']

    return vle['credentials']


def api_set_learner_send(admin, send_category, learner):
    """
        The SEND admin assigns send options to learner
        :param admin: Credentials for a user with SEND administration rights
        :param send_category: Send category object to be assigned to the learner
        :param learner: Learner object
    """
    # Authenticate with admin credentials
    client = auth_utils.client_with_user_credentials(admin['email'], admin['password'])

    # Get the user profile
    profile = auth_utils.get_profile(client)

    # Assign provided SEND category
    institution_id = profile['institution']['id']
    send_cat_resp = client.post(
        '/api/v2/institution/{}/learner/{}/send/'.format(institution_id, learner['id']),
        data={
            'category': send_category['id']
        },
        format='json'
    )
    assert send_cat_resp.status_code == 201



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" InformedConsent model module."""

from cache_memoize import cache_memoize
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .institution import Institution


@cache_memoize(24 * 60 * 60)
def valid_version(version, current_version):
    """
        Check if a version is valid. To be valid, it should be the current version or a minor change
        :param version: The version to be checked
        :type version: str
        :param current_version: The current valid version
        :type current_version: str
        :return: True if the version is valid or False otherwise
        :rtype: bool
    """
    if version == current_version:
        return True
    version = [int(ver_part) for ver_part in version.split('.')]
    current_version = [int(ver_part) for ver_part in current_version.split('.')]
    if version[0] == current_version[0] and version[1] == current_version[1]:
        return True
    return False


@cache_memoize(24 * 60 * 60)
def get_current_ic_version(institution):
    """
        Get the current Informed Version for an institution
        :param institution: Institution object
        :type institution: Institution
        :return: Current version for this institution
        :rtype: str
    """
    institution_ic = InformedConsent.objects.filter(institution=institution,
                                                    valid_from__lte=timezone.now()
                                                    ).all()
    current_version = None
    for inst_ic in institution_ic:
        version = [int(ver_part) for ver_part in inst_ic.version.split('.')]
        if current_version is None:
            current_version = version
        else:
            if version[0] > current_version[0] or (
                version[0] == current_version[0]
                and version[1] > current_version[1]
            ) or (
                version[0] == current_version[0]
                and version[1] == current_version[1]
                and version[2] > current_version[2]
            ):
                current_version = version
    if current_version is not None:
        return '.'.join([str(ver_part) for ver_part in current_version])
    return None


class InformedConsent(BaseModel):
    """ InformedConsent model. """
    institution = models.ForeignKey(Institution, null=False, on_delete=models.CASCADE,
                                    help_text=_('Institution of the informed consent'))
    version = models.CharField(max_length=250, null=False, blank=False,
                               help_text=_("Informed consent version."))

    valid_from = models.DateTimeField(null=False, help_text="Informed consent valid from")

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('institution', 'version'),)

    def __repr__(self):
        return "<InformedConsent(version='%s', valid_from='%r')>" % (
            self.version, self.valid_from)

    @property
    def status(self):
        """
            Informed consent status
            :return: Current status of this informed consent
            :rtype: str
        """
        if self.valid_from > timezone.now():
            return 'NOT_VALID_YET'
        current_version = get_current_ic_version(self.institution)
        if current_version == self.version:
            return 'VALID'
        elif valid_version(self.version, current_version):
            return 'VALID_NEED_UPDATE'
        return 'NOT_VALID'

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        super().save(force_insert, force_update, using, update_fields)
        # Invalidate cached value
        get_current_ic_version.invalidate(self.institution)



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Activity Report api serialize module."""
from rest_framework import serializers

from tesla_ce.models import Learner
from tesla_ce.models import ReportActivity
from tesla_ce.models import ReportActivityInstrument


class InstitutionCourseActivityReportLearnerSerializer(serializers.ModelSerializer):
    """Activity Report Learner serialize model module."""
    class Meta:
        model = Learner
        fields = ["id", "uid", "first_name", "last_name", "email", "learner_id"]


class InstitutionCourseActivityReportDetailSerializer(serializers.ModelSerializer):
    """Activity Report Detail serialize model module."""
    instrument = serializers.CharField(source='instrument.name')
    instrument_id = serializers.IntegerField(source='instrument.id')
    enrolment = serializers.FloatField()
    confidence = serializers.FloatField()
    result = serializers.FloatField()

    class Meta:
        model = ReportActivityInstrument
        fields = ["instrument", "enrolment", "confidence", "result", "identity_level",
                  "content_level", "integrity_level", "instrument_id"]


class InstitutionCourseActivityReportExtendedDetailSerializer(serializers.ModelSerializer):
    """Activity Report Detail serialize model module."""
    instrument = serializers.CharField(source='instrument.name')
    instrument_id = serializers.IntegerField(source='instrument.id')
    instrument_acronym = serializers.CharField(source='instrument.acronym')
    instrument_polarity = serializers.SerializerMethodField()
    enrolment = serializers.FloatField()
    confidence = serializers.FloatField()
    result = serializers.FloatField()
    learner_histogram = serializers.SerializerMethodField()
    activity_histogram = serializers.SerializerMethodField()
    result_bean = serializers.SerializerMethodField()
    thresholds = serializers.SerializerMethodField()
    prob_learner = serializers.SerializerMethodField()
    prob_context = serializers.SerializerMethodField()
    h_prob_learner = serializers.SerializerMethodField()
    h_prob_context = serializers.SerializerMethodField()
    facts = serializers.SerializerMethodField()

    class Meta:
        model = ReportActivityInstrument
        fields = ["instrument", "enrolment", "confidence", "result", "identity_level", "instrument_acronym",
                  "content_level", "integrity_level", "learner_histogram", "activity_histogram",
                  "instrument_id", "instrument_polarity", "result_bean", "thresholds",
                  "prob_learner", "prob_context", "h_prob_learner", "h_prob_context", "facts"]

    @staticmethod
    def get_activity_histogram(instance):
        """
            Get the activity histogram
            :param instance: Report Activity Instrument
            :return: Histogram
        """
        return instance.get_activity_histogram()

    @staticmethod
    def get_learner_histogram(instance):
        """
            Get the learner histogram
            :param instance: Report Activity Instrument
            :return: Histogram
        """
        return instance.get_learner_histogram()

    @staticmethod
    def get_instrument_polarity(instance):
        """
            Return the instrument polarity
            :param instance: Report Activity Instrument
            :return: 1 for normal polarity and -1 if inverted polarity
        """
        return instance.get_instrument_polarity()

    @staticmethod
    def get_result_bean(instance):
        """
            Get the histogram bin where current result is included
            :param instance: Report Activity Instrument
            :return: Bin number
        """
        return instance.get_result_bean()

    @staticmethod
    def get_thresholds(instance):
        """
            Get the reference thresholds for current instrument
            :param instance: Report Activity Instrument
            :return: Reference thresholds
        """
        prov_thr = instance.instrument.provider_set.first()
        return {
            "warning_below": prov_thr.warning_below,
            "alert_below": prov_thr.alert_below
        }

    @staticmethod
    def get_prob_learner(instance):
        """
            Compute the probability to have current value for the learner
            :param instance: Report Activity Instrument
            :return: Probability value
        """
        return instance.get_prob_learner()

    @staticmethod
    def get_prob_context(instance):
        """
            Compute the probability to have current value in the context
            :param instance: Report Activity Instrument
            :return: Probability value
        """
        return instance.get_prob_context()

    @staticmethod
    def get_h_prob_learner(instance):
        """
            Compute the probability to have better value for the learner
            :param instance: Report Activity Instrument
            :return: Probability value
        """
        return instance.get_h_prob_learner()

    @staticmethod
    def get_h_prob_context(instance):
        """
            Compute the probability to have better value in the context
            :param instance: Report Activity Instrument
            :return: Probability value
        """
        return instance.get_h_prob_context()

    @staticmethod
    def get_facts(instance):
        """
            Get facts from results
            :param instance: Report Activity Instrument
            :return: Dictionary with facts
        """
        return instance.get_facts()


class InstitutionCourseActivityReportSerializer(serializers.ModelSerializer):
    """Activity Report serialize model module."""
    learner = InstitutionCourseActivityReportLearnerSerializer()
    detail = InstitutionCourseActivityReportDetailSerializer(source='reportactivityinstrument_set', many=True)

    class Meta:
        model = ReportActivity
        fields = ["id", "learner", "detail", "identity_level", "content_level", "integrity_level"]


class InstitutionCourseActivityReportExtendedSerializer(serializers.ModelSerializer):
    """Activity Report serialize model module."""
    learner = InstitutionCourseActivityReportLearnerSerializer()
    detail = InstitutionCourseActivityReportExtendedDetailSerializer(source='reportactivityinstrument_set', many=True)
    data = serializers.FileField(read_only=True)

    class Meta:
        model = ReportActivity
        fields = ["id", "learner", "detail", "identity_level", "content_level", "integrity_level", "data"]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Course from a Course Group views module """
from django.db.models import Q
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import InstitutionCourseSerializer
from tesla_ce.models import Course
from tesla_ce.models.user import is_global_admin
from tesla_ce.models.user import get_institution_user


# pylint: disable=too-many-ancestors
class InstitutionCourseViewSet(NestedViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    API endpoint that allows course in a vle to be viewed or edited.
    """
    model = Course
    serializer_class = InstitutionCourseSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    permission_classes = [
        permissions.InstitutionMemberReadOnlyPermission |
        permissions.GlobalAdminReadOnlyPermission |
        permissions.InstitutionAdminPermission
    ]
    # filterset_fields = ['code', 'vle_course_id', 'description', 'vle']
    # search_fields = ['vle_course_id', 'description', 'code', 'vle']

    def get_queryset(self):
        # Except for global admins, and institution admins and data admins, returns only the courses where the user is
        # involved as instructor or learner
        qs = super().filter_queryset_by_parents_lookups(Course.objects)
        if not is_global_admin(self.request.user):
            inst_user = get_institution_user(self.request.user)
            if not inst_user.inst_admin and not inst_user.data_admin:
                qs = qs.filter(
                    Q(instructors__id=inst_user.id) | Q(learners__id=inst_user.id)
                ).distinct()
        return qs.all()



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Launcher model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .assessment_session import AssessmentSession
from .base_model import BaseModel
from .user import InstitutionUser

LAUNCHER_TARGET = (
    (0, _('Dashboard')),
    (1, _('LAPI')),
)


class Launcher(BaseModel):
    """ Launcher model. """
    user = models.ForeignKey(InstitutionUser, null=False,
                             on_delete=models.CASCADE,
                             help_text=_("Related user"))

    session = models.ForeignKey(AssessmentSession, null=True, default=None,
                                on_delete=models.SET_NULL,
                                help_text=_("Related assessment session"))

    target = models.IntegerField(choices=LAUNCHER_TARGET, null=False,
                                 default=0,
                                 help_text=_('Target for this launcher'))

    target_url = models.CharField(max_length=250, null=True, default=None,
                                  help_text=_('Url to redirect'))

    token = models.UUIDField(null=False, help_text=_("Token to allow authentication for this user"))

    token_pair = models.TextField(null=False, help_text=_('Pair of tokens to authenticate with target'))

    expires_at = models.DateTimeField(null=False,
                                      help_text=_("When this access token expires"))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE API global tests """
import pytest
import json


@pytest.mark.django_db
def test_favicon(rest_api_client):
    favicon_resp = rest_api_client.get('/favicon.ico')
    assert favicon_resp.status_code == 302


@pytest.mark.django_db
def test_api_version(rest_api_client):
    api_version_resp = rest_api_client.get('/api/version/')
    assert api_version_resp.status_code == 200
    assert 'version' in json.loads(api_version_resp.content)

    lapi_version_resp = rest_api_client.get('/lapi/version/')
    assert lapi_version_resp.status_code == 200
    assert 'version' in json.loads(lapi_version_resp.content)

    assert json.loads(api_version_resp.content)['version'] == json.loads(lapi_version_resp.content)['version']





#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Instructor actions for Use Case tests """
import requests
from tests import auth_utils


def api_configure_activity(launcher, activity, providers):
    """
        An instructor configures the activity using the API
        :param launcher: Launcher object for an instructor
        :param activity: Activity object
        :param providers: List of enabled providers
    """
    # Authenticate with instructor launcher credentials
    client = auth_utils.client_with_launcher_credentials(launcher)

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "INSTRUCTOR" in profile['roles']

    # Get required data
    institution_id = profile['institution']['id']
    course_id = activity['course']['id']
    activity_id = activity['id']

    # Add KS as main instrument
    instrument_id = providers['ks']['instrument']['id']
    ks_data = {
        'active': True,
        'instrument_id': instrument_id
    }
    ks_add_response = client.post(
        '/api/v2/institution/{}/course/{}/activity/{}/instrument/'.format(institution_id, course_id, activity_id),
        data=ks_data,
        format='json'
    )
    assert ks_add_response.status_code == 201
    # Add FR as an alternative instrument to KS
    instrument_id = providers['fr']['instrument']['id']
    fr_data = {
        'active': True,
        'options': {'online': True},
        'instrument_id': instrument_id,
        'alternative_to': ks_add_response.data['id']
    }
    fr_add_response = client.post(
        '/api/v2/institution/{}/course/{}/activity/{}/instrument/'.format(institution_id, course_id, activity_id),
        data=fr_data,
        format='json'
    )
    assert fr_add_response.status_code == 201
    # Add plagiarism as main instrument
    instrument_id = providers['plag']['instrument']['id']
    plag_data = {
        'active': True,
        'instrument_id': instrument_id
    }
    plag_add_response = client.post(
        '/api/v2/institution/{}/course/{}/activity/{}/instrument/'.format(institution_id, course_id, activity_id),
        data=plag_data,
        format='json'
    )
    assert plag_add_response.status_code == 201

    # Check that activity has now 3 instruments
    inst_list_response = client.get(
        '/api/v2/institution/{}/course/{}/activity/{}/instrument/'.format(institution_id, course_id, activity_id)
    )
    assert inst_list_response.status_code == 200
    assert inst_list_response.data['count'] == 3


def api_instructor_report(launcher, activity, filters=None):
    """
        Instructor review the results for the activity
        :param launcher: Instructor launcher object
        :param activity: Activity object
        :param filters: Filter reports
    """
    # Authenticate with instructor launcher credentials
    client = auth_utils.client_with_launcher_credentials(launcher)

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "INSTRUCTOR" in profile['roles']

    # Get required data
    institution_id = profile['institution']['id']
    course_id = activity['course']['id']
    activity_id = activity['id']

    # Apply filters if provided
    url = '/api/v2/institution/{}/course/{}/activity/{}/report/'
    if filters is not None:
        url += '?{}'.format(filters)

    # Get the list of reports from the activity
    activity_reports_resp = client.get(url.format(
        institution_id,
        course_id,
        activity_id
    ))
    assert activity_reports_resp.status_code == 200
    reports = activity_reports_resp.data['results']

    # Get the detail for each report
    for report in reports:
        reports_detail_resp = client.get('/api/v2/institution/{}/course/{}/activity/{}/report/{}/'.format(
            institution_id,
            course_id,
            activity_id,
            report['id']
        ))
        assert reports_detail_resp.status_code == 200
        report['detailed_report'] = reports_detail_resp.data

        report_data_resp = requests.get(report['detailed_report']['data'], verify=False)
        assert report_data_resp.status_code == 200
        report['detailed_report']['data'] = report_data_resp.json()
        assert 'data' in report['detailed_report']
        assert 'sessions' in report['detailed_report']['data']

        request_list = []
        final = False
        while not final:
            report_requests = client.get(
                '/api/v2/institution/{}/course/{}/activity/{}/report/{}/request/?offset={}'.format(
                    institution_id,
                    course_id,
                    activity_id,
                    report['id'],
                    len(request_list)
            ))
            assert report_requests.status_code == 200
            request_list += report_requests.data['results']
            if len(request_list) >= report_requests.data['count']:
                final = True
        report['detailed_report']['requests'] = request_list

        # Get audit data for each instrument
        report['audit'] = {}
        for instrument in report['detail']:
            audit_resp = client.get(
                '/api/v2/institution/{}/course/{}/activity/{}/report/{}/audit/{}/'.format(
                    institution_id,
                    course_id,
                    activity_id,
                    report['id'],
                    instrument['instrument_id']
            ))
            assert audit_resp.status_code == 200
            report['audit'][instrument['instrument_id']] = audit_resp.data

    return reports



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for users administration """
import pytest

from tests.conftest import get_random_string

from tests import auth_utils


@pytest.mark.django_db
def test_api_admin_user(rest_api_client, user_global_admin):
    """
        CRUD for users without institution
    """
    # Set global admin user.
    client = auth_utils.client_with_user_obj(user_global_admin)

    # Create a new user
    user_name = get_random_string(10)
    password = get_random_string(10)
    email = '{}@tesla-ce.eu'.format(user_name)
    user_data = {
        'username': user_name,
        'uid': user_name,
        'email': email,
        'first_name': user_name[:5],
        'last_name': user_name[5:],
        'login_allowed': True,
        'is_active': True,
        'is_staff': False,
        'password': password,
        'password2': password
    }
    user_create_resp = client.post('/api/v2/admin/user/', data=user_data, format='json')
    assert user_create_resp.status_code == 201

    # Ensure that this user cannot authenticate, as is not in any institution and is not an GLOBAL_ADMIN
    try:
        auth_utils.client_with_user_credentials(email, password)
        pytest.fail('User without institution and admin privileges should not be able to login')
    except AssertionError:
        pass

    # Grant administration privileges
    user_mod_resp = client.patch(
        '/api/v2/admin/user/{}/'.format(user_create_resp.data['id']),
        data={'is_staff': True},
        format='json'
    )
    assert user_mod_resp.status_code == 200

    # Get the user profile
    user_client = auth_utils.client_with_user_credentials(email, password)
    profile = auth_utils.get_profile(user_client)

    assert len(profile['roles']) == 1
    assert 'GLOBAL_ADMIN' in profile['roles']
    assert profile['username'] == user_name
    assert profile['email'] == email
    assert profile['first_name'] == user_name[:5]
    assert profile['last_name'] == user_name[5:]

    # Find for this user by email
    find_user_resp = client.get('/api/v2/admin/user/?email={}'.format(email))
    assert find_user_resp.status_code == 200
    assert find_user_resp.data['count'] == 1
    assert find_user_resp.data['results'][0]['id'] == profile['id']

    # Remove the user
    user_del_resp = client.delete(
        '/api/v2/admin/user/{}/'.format(user_create_resp.data['id'])
    )
    assert user_del_resp.status_code == 204

    # Ensure that this user does not exist
    find2_user_resp = client.get('/api/v2/admin/user/?email={}'.format(email))
    assert find2_user_resp.status_code == 200
    assert find2_user_resp.data['count'] == 0


@pytest.mark.django_db
def test_api_admin_institution_user(rest_api_client, user_global_admin, institution_test_case):
    """
        CRUD for users with institution
    """
    # Set global admin user.
    client = auth_utils.client_with_user_obj(user_global_admin)

    # Create a new user
    user_name = get_random_string(10)
    password = get_random_string(10)
    email = '{}@tesla-ce.eu'.format(user_name)
    user_data = {
        'username': user_name,
        'uid': user_name,
        'email': email,
        'first_name': user_name[:5],
        'last_name': user_name[5:],
        'login_allowed': True,
        'is_active': True,
        'is_staff': False,
        'institution_id': institution_test_case['institution'].id,
        'password': password,
        'password2': password
    }
    user_create_resp = client.post('/api/v2/admin/user/', data=user_data, format='json')
    assert user_create_resp.status_code == 201

    # Get the user profile
    user_client = auth_utils.client_with_user_credentials(email, password)
    profile = auth_utils.get_profile(user_client)

    assert len(profile['roles']) == 0
    assert 'institution' in profile
    assert profile['institution']['id'] == institution_test_case['institution'].id
    assert profile['email'] == email
    assert profile['username'] == user_name
    assert profile['last_name'] == user_name[5:]
    assert profile['first_name'] == user_name[:5]

    # Find for this user by email
    find_user_resp = client.get('/api/v2/admin/user/?email={}'.format(email))
    assert find_user_resp.status_code == 200
    assert find_user_resp.data['count'] == 1
    assert find_user_resp.data['results'][0]['id'] == profile['id']

    # Set administration rights on the institution
    mod_user_resp = client.patch('/api/v2/admin/user/{}/'.format(profile['id']),
                                 data={'inst_admin': True}, format='json')
    assert mod_user_resp.status_code == 200

    # Get user data
    user_data_resp = client.get(
        '/api/v2/admin/user/{}/'.format(user_create_resp.data['id'])
    )
    assert user_data_resp.status_code == 200
    assert len(user_data_resp.data['roles']) == 1
    assert "ADMIN" in user_data_resp.data['roles']

    # Remove the user
    user_del_resp = client.delete(
        '/api/v2/admin/user/{}/'.format(user_create_resp.data['id'])
    )
    assert user_del_resp.status_code == 204

    # Ensure that this user does not exist
    find2_user_resp = client.get('/api/v2/admin/user/?email={}'.format(email))
    assert find2_user_resp.status_code == 200
    assert find2_user_resp.data['count'] == 0


@pytest.mark.django_db
def test_api_admin_user_mix(rest_api_client, user_global_admin, institution_test_case):
    """
        Test the transformation of users from institution to global and opposite
    """
    # Set global admin user.
    client = auth_utils.client_with_user_obj(user_global_admin)

    # Create a new user without institution
    user_name = get_random_string(10)
    password = get_random_string(10)
    email = '{}@tesla-ce.eu'.format(user_name)
    user_data = {
        'username': user_name,
        'uid': user_name,
        'email': email,
        'first_name': user_name[:5],
        'last_name': user_name[5:],
        'login_allowed': True,
        'is_active': True,
        'is_staff': False,
        'password': password,
        'password2': password
    }
    user_create_resp = client.post('/api/v2/admin/user/', data=user_data, format='json')
    assert user_create_resp.status_code == 201
    assert user_create_resp.data['institution'] is None

    # Assign the user to an institution
    user_mod_resp = client.patch(
        '/api/v2/admin/user/{}/'.format(user_create_resp.data['id']),
        data={
            'institution_id': institution_test_case['institution'].id,
            'uid': email
        },
        format='json'
    )
    assert user_mod_resp.status_code == 200
    assert user_mod_resp.data['institution'] is not None
    assert user_mod_resp.data['institution']['id'] == institution_test_case['institution'].id
    assert user_mod_resp.data['email'] == email
    assert user_mod_resp.data['first_name'] == user_name[:5]
    assert user_mod_resp.data['last_name'] == user_name[5:]

    # Create a new institution
    new_inst_resp = client.post('/api/v2/admin/institution/', data={
        'name': "PyTest Test institution",
        'acronym': get_random_string(10)
    }, format='json')
    assert new_inst_resp.status_code == 201

    # Assign the user to an institution
    user_mod2_resp = client.patch(
        '/api/v2/admin/user/{}/'.format(user_create_resp.data['id']),
        data={'institution_id': new_inst_resp.data['id']},
        format='json'
    )
    assert user_mod2_resp.status_code == 200
    assert user_mod2_resp.data['institution'] is not None
    assert user_mod2_resp.data['institution']['id'] == new_inst_resp.data['id']
    assert user_mod2_resp.data['email'] == email
    assert user_mod2_resp.data['first_name'] == user_name[:5]
    assert user_mod2_resp.data['last_name'] == user_name[5:]

    # Remove the institution from user
    user_mod3_resp = client.patch(
        '/api/v2/admin/user/{}/'.format(user_create_resp.data['id']),
        data={'institution_id': -1},
        format='json'
    )
    assert user_mod3_resp.status_code == 200
    assert user_mod3_resp.data['institution'] is None
    assert user_mod3_resp.data['email'] == email
    assert user_mod3_resp.data['first_name'] == user_name[:5]
    assert user_mod3_resp.data['last_name'] == user_name[5:]


@pytest.mark.django_db
def test_api_admin_user_filters(rest_api_client, user_global_admin, institution_test_case):
    """
        Test filters for getting the users
    """
    # Set global admin user.
    client = auth_utils.client_with_user_obj(user_global_admin)

    # Get the global admin user by email
    get_user_by_email_resp = client.get('/api/v2/admin/user/?email={}'.format(user_global_admin.email))
    assert get_user_by_email_resp.status_code == 200
    assert get_user_by_email_resp.data['count'] == 1
    assert get_user_by_email_resp.data['results'][0]['id'] == user_global_admin.id

    # Filter users using one role
    found = False
    finalize = False
    offset = 0
    while not finalize:
        get_user_by_role_resp = client.get('/api/v2/admin/user/?offset={}&roles=GLOBAL_ADMIN'.format(offset))
        assert get_user_by_role_resp.status_code == 200
        for user in get_user_by_role_resp.data['results']:
            assert 'GLOBAL_ADMIN' in user['roles']
            if user['id'] == user_global_admin.id:
                found = True
                finalize = True
        if len(get_user_by_role_resp.data['results']) == 0:
            finalize = True
        offset += len(get_user_by_role_resp.data['results'])
    assert found

    # Filter users with multiple roles
    found1 = False
    found2 = False
    finalize = False
    offset = 0
    inst_user = institution_test_case['user'].institutionuser
    inst_user.legal_admin = True
    inst_user.save()
    while not finalize:
        get_user_by_roles_resp = client.get(
            '/api/v2/admin/user/?offset={}&roles=GLOBAL_ADMIN&roles=LEGAL'.format(offset)
        )
        assert get_user_by_roles_resp.status_code == 200
        for user in get_user_by_roles_resp.data['results']:
            assert 'GLOBAL_ADMIN' in user['roles'] or 'LEGAL' in user['roles']
            if user['id'] == user_global_admin.id:
                found1 = True
            if user['id'] == inst_user.id:
                found2 = True
            if found1 and found2:
                finalize = True
        if len(get_user_by_roles_resp.data['results']) == 0:
            finalize = True
        offset += len(get_user_by_roles_resp.data['results'])
    assert found1 and found2

    # Filter users by institution
    get_user_by_institution_resp = client.get(
        '/api/v2/admin/user/?institution={}'.format(institution_test_case['institution'].id)
    )
    assert get_user_by_institution_resp.status_code == 200
    assert get_user_by_institution_resp.data['count'] == 1
    assert get_user_by_institution_resp.data['results'][0]['id'] == inst_user.id

    # Filter users by passing no institution (all users expected)
    get_user_by_all_institution_resp = client.get(
        '/api/v2/admin/user/?institution='
    )
    assert get_user_by_all_institution_resp.status_code == 200
    assert get_user_by_all_institution_resp.data['count'] > 1

    # Filter users with no assignedinstitution
    get_user_by_no_institution_resp = client.get(
        '/api/v2/admin/user/?institution=-1'
    )
    assert get_user_by_no_institution_resp.status_code == 200
    assert get_user_by_no_institution_resp.data['count'] > 0
    for user in get_user_by_no_institution_resp.data['results']:
        assert user['institution'] is None



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" VLE model module."""
from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .institution import Institution

# VLE types
VLE_TYPE = (
    (0, _('MOODLE')),
)


def get_type_id(name):
    """
        Get the vle type ID from vle name
        :param name: VLE type name
        :return: Vle type ID
    """
    for vle_type in VLE_TYPE:
        if vle_type[1].upper() == name.upper():
            return vle_type[0]
    return None


class VLE(BaseModel):
    """ VLE model. """
    name = models.CharField(max_length=250, null=False, blank=False, unique=True,
                            help_text=_("VLE unique name."))
    type = models.IntegerField(choices=VLE_TYPE, null=False, default=0, help_text=_('Type of the VLE.'))
    institution = models.ForeignKey(Institution, null=False, on_delete=models.CASCADE,
                                    help_text=_('Institution this VLE belongs to.'))
    url = models.TextField(null=True, blank=True, help_text=_("VLE url."))
    lti = models.TextField(null=True, blank=True, help_text=_("VLE lti configuration."))
    client_id = models.CharField(max_length=250, null=True, blank=False, default=None,
                                 help_text=_("LTI 1.3 Client ID."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<VLE(id='%r', type='%r', name='%s')>" % (
            self.id, self.get_type_display(), self.name)



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Report Activity model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .activity import Activity
from .base_model import BaseModel
from .learner import Learner

REPORT_ALERT_LEVEL = (
    (0, _('Pending')),
    (1, _('No Information')),
    (2, _('Ok')),
    (3, _('Warning')),
    (4, _('Alert')),
)


class ReportActivity(BaseModel):
    """ Report Activity model. """

    learner = models.ForeignKey(Learner, null=False, blank=False, on_delete=models.CASCADE,
                                help_text=_('Learner related to this report.'))

    activity = models.ForeignKey(Activity, null=False, blank=False, on_delete=models.CASCADE,
                                 help_text=_('Activity related to this report.'))

    identity_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                              help_text=_('Alert level for learner identity.'))
    content_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                             help_text=_('Alert level for content authorship.'))
    integrity_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                               help_text=_('Alert level for system integrity.'))

    data = models.FileField(max_length=250, null=True, blank=False,
                            help_text=_("Path to the content of this report."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('activity', 'learner'),)

    def __repr__(self):
        return "<ReportActivity(id='%r', learner_id='%r' activity_id='%r')>" % (
            self.id, self.learner_id, self.activity_id)



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Activity model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .course import Course
from .vle import VLE


class Activity(BaseModel):
    """ Activity model. """
    vle = models.ForeignKey(VLE, null=False, blank=False,
                            on_delete=models.CASCADE,
                            help_text=_('Activity VLE.'))

    course = models.ForeignKey(Course, null=False, blank=False,
                            on_delete=models.CASCADE,
                            help_text=_('Activity VLE.'))

    vle_activity_type = models.CharField(max_length=250, null=False, blank=False,
                                         help_text=_("Activity type on the VLE."))

    vle_activity_id = models.CharField(max_length=250, null=False, blank=False,
                                       help_text=_("Activity id on the VLE."))

    description = models.TextField(null=True, blank=True,
                                   help_text=_("Activity description."))

    name = models.CharField(max_length=255, null=True, blank=True,
                            help_text=_("Activity name."))

    enabled = models.BooleanField(null=None, blank=None, default=False,
                                  help_text=_("Whether this activity is enabled or not"))

    start = models.DateTimeField(blank=True, null=True,
                                 help_text=_("When activity starts"))

    end = models.DateTimeField(blank=True, null=True,
                               help_text=_("When activity ends"))

    conf = models.TextField(null=True, blank=True,
                            help_text=_("Activity conf."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('vle', 'vle_activity_type', 'vle_activity_id'),)

    def __repr__(self):
        return "<Activity(id='%r', course_id='%r' vle_id='%r', vle_act_type='%s', vle_act_id='%r', enabled='%r')>" % (
            self.id, self.course_id, self.course.vle_id, self.vle_activity_type, self.vle_activity_id, self.enabled)

    def get_learner_instruments(self, learner):
        """
        Get the list of instruments that should be activated for this activity for a given learner.

        :param learner: The learner identifier
        :type learner: uuid

        :return: List of instrument configurations for this activity
        """
        disabled_instruments = []
        if learner.send['is_send'] and 'disabled_instruments' in learner.send['send']:
            disabled_instruments = learner.send['send']['disabled_instruments']

        primary_instruments = self.configuration.filter(active=True,
                                                        alternative_to=None
                                                        ).exclude(instrument_id__in=disabled_instruments).all()
        excluded_instruments = self.configuration.filter(active=True,
                                                         alternative_to=None,
                                                         instrument_id__in=disabled_instruments).all()
        alternative_instruments = []
        for exc_inst in excluded_instruments:
            alternative = exc_inst.activityinstrument_set.filter(active=True).all()
            alternative_instruments += list(alternative)

        return list(primary_instruments) + alternative_instruments



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA Authentication views"""
import hvac
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db import transaction
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.views import APIView
from rest_framework.views import Response
from rest_framework.views import status

from tesla_ce import models
from tesla_ce.apps.api.utils import decode_json
from tesla_ce.client import Client
from tesla_ce.lib.auth import JWTAuthentication
from tesla_ce.lib.auth import _get_token
from tesla_ce.lib.exception import TeslaAuthException
from . import serializers


class AuthSchema(AutoSchema):
    """ Schema definition for Authentication methods """
    def get_operation(self, path, method):
        default = super().get_operation(path, method)
        if path.endswith('user'):
            default['requestBody'] = {
                'content': {
                    'application/json': {
                        'schema': {
                            'type': 'object',
                            'required': ['email', 'password'],
                            'properties': {
                                'email': {
                                    'type': 'string',
                                },
                                'password': {
                                    'type': 'string',
                                },
                                'realm': {
                                    'type': 'string',
                                },
                            }
                        }
                    }
                }
            }
        elif path.endswith('approle'):
            default['requestBody'] = {
                'content': {
                    'application/json': {
                        'schema': {
                            'type': 'object',
                            'required': ['role_id', 'secret_id'],
                            'properties': {
                                'role_id': {
                                    'type': 'string',
                                },
                                'secret_id': {
                                    'type': 'string',
                                },
                            }
                        }
                    }
                }
            }
        elif path.endswith('token'):
            default['requestBody'] = {
                'content': {
                    'application/json': {
                        'schema': {
                            'type': 'object',
                            'required': ['token'],
                            'properties': {
                                'token': {
                                    'type': 'string',
                                }
                            }
                        }
                    }
                }
            }
        elif path.endswith('token/refresh'):
            default['requestBody'] = {
                'content': {
                    'application/json': {
                        'schema': {
                            'type': 'object',
                            'required': ['refresh_token'],
                            'properties': {
                                'refresh_token': {
                                    'type': 'string',
                                }
                            }
                        }
                    }
                }
            }

        return default


class AuthViewBase(APIView):
    """
        Base authentication view
    """
    # API TeSLA Client
    _client = Client(config=settings.TESLA_CONFIG)

    # Vault client to manage authentication requests
    _vault = hvac.Client(settings.TESLA_CONFIG.config.get('VAULT_URL'),
                         verify=settings.TESLA_CONFIG.config.get('VAULT_SSL_VERIFY'))

    permission_classes = ()
    authentication_classes = ()
    schema = AuthSchema()

    # Serializer for request body
    serializer_class = None
    alternative_serializer_class = None

    def get_authenticate_header(self, request):
        """
            Get the name of authentication token
            :param request: The HTTP request
            :return: Name of the token
        """
        return 'JWT'

    def authenticate(self, data):
        """
            Perform authentication using provided data

            :param data: Authentication data
            :return: Authenticated object or None if is not authenticated
        """
        return None

    @property
    def get_serializer(self):
        """
            Get the serializer for authentication object
            :return: Serializer class
            :rtype: class or None
        """
        return self.serializer_class

    @property
    def get_alternative_serializer(self):
        """
            Get alternative serializer for authentication class.
            :return: Serializer class
            :rtype: class or None
        """
        return self.alternative_serializer_class

    def get_entity_config(self, entity_id):
        """
            Get entity configuration data
            :param entity_id: Entity ID
            :return: Configuration data
        """
        entity_data = self._vault.secrets.kv.v2.read_secret_version(
            path='modules/{}/metadata'.format(entity_id),
            mount_point=self._client.config.config.get('VAULT_MOUNT_PATH_KV')
        )['data']['data']['data']
        entity_data['apps'] = self._vault.secrets.kv.v2.read_secret_version(
            path=entity_data['apps'],
            mount_point=self._client.config.config.get('VAULT_MOUNT_PATH_KV')
        )['data']['data']['data']
        entity_data['config'] = self._vault.secrets.kv.v2.read_secret_version(
            path=entity_data['config'],
            mount_point=self._client.config.config.get('VAULT_MOUNT_PATH_KV')
        )['data']['data']['data']
        entity_data['services'] = self._vault.secrets.kv.v2.read_secret_version(
            path=entity_data['services'],
            mount_point=self._client.config.config.get('VAULT_MOUNT_PATH_KV')
        )['data']['data']['data']

        config = {}
        for key, path in entity_data['config']:
            config[key] = self._vault.secrets.kv.v2.read_secret_version(
                path=path,
                mount_point=self._client.config.config.get('VAULT_MOUNT_PATH_KV')
            )['data']['data']['value']
        entity_data['config'] = config

        return entity_data

    def post(self, request, *args, **kwargs):
        """
            Dispatcher for POST HTTP method
            :param request: The HTTP request
            :param args: Arguments provided to the request
            :param kwargs: Named arguments provided to the request
            :return: HTTP response
        """
        serializer = self.get_serializer(data=request.data)

        if not serializer.is_valid():
            if self.get_alternative_serializer is None:
                return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
            serializer = self.get_alternative_serializer(data=request.data)
            if not serializer.is_valid():
                return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        auth_data = self.authenticate(serializer.validated_data)
        if auth_data is None:
            return Response("Invalid credentials", status=status.HTTP_401_UNAUTHORIZED)

        # Add tokens
        auth_data = self._add_token(auth_data)

        return Response(auth_data, status=status.HTTP_200_OK)

    def get(self, request):
        """
            Dispatcher for GET HTTP method
            :param request: The HTTP request
            :return: HTTP response
        """
        return Response('Method not allowed', status=status.HTTP_405_METHOD_NOT_ALLOWED)

    def _add_token(self, auth_data):
        """
            Add tokens for given authentication data

            :param auth_data: Authentication object
            :type auth_data: dict
            :return: Authenticated object with token information
            :rtype: dict
        """
        token = self._client.get_module_token(module_id='api', scope=['/api/*'],
                                              ttl=15, data={
                'type': auth_data['type'],
                'pk': auth_data['pk'],
            })

        refresh_token = self._client.get_module_token(module_id='api', scope=['/api/v2/auth/token/refresh'],
                                                      ttl=120,
                                                      data={
                                                          'ttl': 15,
                                                          'type': auth_data['type'],
                                                          'pk': auth_data['pk'],
                                                      }
                                                      )
        auth_data['token'] = {
            'access_token': token,
            'refresh_token': refresh_token
        }
        return auth_data


class UserPasswordView(AuthViewBase):
    """
        Authentication View for user/password credentials
    """
    serializer_class = serializers.AuthUserPasswordSerializer

    def authenticate(self, data):
        """
            Verify user credentials
            :param data: Provided credentials with email and password fields
            :return: Authenticated object or None if authentication fails
        """
        try:
            return self._client.verify_user(data['email'], data['password'])
        except TeslaAuthException:
            return None

    def _add_token(self, auth_data):
        return {
            'token': auth_data
        }


class AppRoleView(AuthViewBase):
    """
        Authentication View for role_id/secret_id credentials
    """
    serializer_class = serializers.AppRoleTokenSerializer

    def authenticate(self, data):
        """
            Verify module credentials
            :param data: Provided credentials with role_id and secret_id fields
            :return: Authenticated object or None if authentication fails
        """

        # Authenticate using approle credentials
        try:
            auth_resp = self._vault.auth.approle.login(
                data['role_id'],
                data['secret_id'],
                mount_point=self._client.config.config.get('VAULT_MOUNT_PATH_APPROLE')
            )
        except Exception:
            return None

        # Get the entity data
        entity_data = self.get_entity_config(auth_resp['auth']['metadata']['role_name'])

        # Get the module object
        if entity_data['module'].startswith('vle_'):
            entity_data['type'] = 'vle'
            entity_data['pk'] = entity_data['vle_id']
            # Get VLE information
            try:
                vle = models.VLE.objects.get(pk=entity_data['vle_id'])
            except models.VLE.DoesNotExist:
                return None
            entity_data['vle'] = serializers.VLESerializer(instance=vle).data

        elif entity_data['module'].startswith('provider_'):
            entity_data['type'] = 'provider'
            entity_data['pk'] = entity_data['provider_id']
            # Get Provider information
            try:
                provider = models.Provider.objects.get(pk=entity_data['provider_id'])
            except models.Provider.DoesNotExist:
                return None
            entity_data['provider'] = serializers.ProviderSerializer(instance=provider).data
        else:
            return None

        return entity_data


class TokenObtainPairView(AuthViewBase):
    """
        Token based authentication view
    """
    serializer_class = serializers.AuthTokenSerializer

    def authenticate(self, data):
        """
            Verify token
            :param data: Provided credentials with id and token fields
            :return: Authenticated object or None if authentication fails
        """
        with transaction.atomic():
            try:
                launcher = models.Launcher.objects.get(id=data['id'],
                                                       token=data['token'])
            except models.Launcher.DoesNotExist:
                return None
        return launcher.token_pair

    def _add_token(self, auth_data):
        return decode_json(auth_data)


class TokenRefreshView(AuthViewBase):
    """
        Refresh token view
    """
    serializer_class = serializers.AuthTokenRefreshSerializer
    alternative_serializer_class = serializers.AlternativeAuthTokenRefreshSerializer

    def authenticate(self, data):
        """
            Verify credentials
            :param data: Provided credentials
            :return: Authenticated object or None if authentication fails
        """
        return data

    def _add_token(self, auth_data):
        token = auth_data['token']
        if 'access_token' in token and 'refresh_token' in token:
            access_token = token['access_token']
            refresh_token = token['refresh_token']
        else:
            access_token = token
            refresh_token = _get_token(self.request)
        try:
            access_token = self._client.refresh_token(access_token, refresh_token)
        except TeslaAuthException as exc:
            raise PermissionDenied(exc) from exc
        return {
            "token": {
                "access_token": access_token,
                "refresh_token": refresh_token
            }
        }


class LogOut(APIView):
    """
        Perform system log out
    """
    def delete(self, request):
        """
            Dispatcher for delete HTTP method
            :param request: The HTTP request
            :return: A HTTP response
        """
        return Response({}, status=status.HTTP_204_NO_CONTENT)


class UserDataView(APIView):
    """
        Get user data
    """
    permission_classes = ()
    authentication_classes = ()
    _authenticator = JWTAuthentication()

    def get(self, request):
        """
            Dispatcher for GET HTTP method

            :param request: The HTTP request
            :return: A HTTP response
        """
        try:
            user, _ = self._authenticator.authenticate(request)
        except TeslaAuthException:
            return Response(status=status.HTTP_401_UNAUTHORIZED)
        if not user.is_authenticated:
            return Response(status=status.HTTP_401_UNAUTHORIZED)
        if user.id is None:
            return Response(status=status.HTTP_403_FORBIDDEN)

        serializer = serializers.UserDataSerializer(user)
        return Response(serializer.data, status=status.HTTP_200_OK)


user_view = UserPasswordView.as_view()
app_view = AppRoleView.as_view()
token_view = TokenObtainPairView.as_view()
token_refresh_view = TokenRefreshView.as_view()
user_data_view = UserDataView.as_view()
logout_view = LogOut.as_view()



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA Client module"""
import json
import os
import uuid

import simplejson
from cache_memoize import cache_memoize
from django.conf import settings
from django.core.files.base import ContentFile
from django.template.loader import render_to_string
from django.templatetags.static import static
from django.utils import timezone

from . import models
from .lib import ConfigManager
from .lib import DatabaseManager
from .lib import DeploymentManager
from .lib import StorageManager
from .lib import VaultManager
from .lib.checks import check_db_configuration
from .lib.checks import check_tesla_configuration
from .lib.checks import check_vault_configuration
from .lib.decorators import tesla_mode_required
from .lib.exception import TeslaAuthException
from .lib.exception import TeslaConfigException
from .lib.exception import TeslaInvalidICException
from .lib.exception import TeslaMissingEnrolmentException
from .lib.exception import TeslaMissingICException
from .lib.exception import TeslaVaultException

#: Client instance
_client = None


def get_version():
    """
        Get current version
        :return: Version value
    """
    version_file = os.path.abspath(os.path.join(os.path.dirname(__file__), 'lib', 'data', 'VERSION'))
    with open(version_file, 'r') as v_file:
        version = v_file.read()
    version = version.strip()
    return version


def get_default_client():
    """
        Get a client instance

        :return: Client instance
        :rtype: Client
    """
    global _client

    if _client is not None:
        return _client
    if not hasattr(settings, 'TESLA_CONFIG'):
        raise TeslaConfigException('Missing TeSLA configuration.')
    if len(settings.TESLA_CONFIG.enabled_modules) == 0:
        raise TeslaConfigException('No module is active.')

    # Get the most appropriated configuration
    config = None
    if len(settings.TESLA_CONFIG.enabled_modules) == 1:
        config = settings.TESLA_CONFIG.get_module_config(settings.TESLA_CONFIG.enabled_modules[0])
    elif 'api' in settings.TESLA_CONFIG.enabled_modules:
        config = settings.TESLA_CONFIG.get_module_config('api')
    elif 'lapi' in settings.TESLA_CONFIG.enabled_modules:
        config = settings.TESLA_CONFIG.get_module_config('lapi')
    else:
        config = settings.TESLA_CONFIG.get_module_config(settings.TESLA_CONFIG.enabled_modules[0])

    # Create the client instance
    _client = Client.create_instance(config['VAULT_URL'], config['VAULT_ROLE_ID'],
                                     config['VAULT_SECRET_ID'], config['VAULT_SSL_VERIFY'])
    return _client


class Client():
    """
        TeSLA Client, that provides a high level interface with the system

        :param config: Access to the configuration manager
        :type config: class:`tesla_ce.lib.config.ConfigManager`
        :param management_mode: Indicates whether the client is in management mode
        :type management_mode: bool
    """
    # Client version
    version = get_version()

    @staticmethod
    def create_instance(vault_url, role_id, secret_id, verify_ssl=True, approle_path=None, kv_path=None):
        """
            Create a client instance from Vault configuration

            :param vault_url: Url to connect with Vault.
            :type vault_url: str
            :param role_id: RoleID value.
            :type role_id: str
            :param secret_id: SecretID value.
            :type secret_id: str
            :param verify_ssl: Verify Vault certificate
            :type verify_ssl: bool
            :param approle_path: Mount point for AppRole authentication service
            :type approle_path: str
            :param kv_path: Mount point for KV secrets service
            :type kv_path: str
            :return: Client instance
            :rtype: Client
        """
        config = ConfigManager(load_config=False)
        config.config.set('VAULT_SSL_VERIFY', verify_ssl)
        config.load_vault(vault_url=vault_url, role_id=role_id, secret_id=secret_id,
                          kv_path=kv_path, approle_path=approle_path)
        return Client(config=config)

    def __init__(self, config=None, config_file=None, use_vault=True, use_env=True, enable_management=False):
        """
            Default constructor from configuration options

            :param config_file: Configuration file
            :type config_file: str
            :param use_vault: User vault to retrieve configuration values
            :type use_vault: bool
            :param use_env: User environment variables and secrets
            :type use_env: bool
            :param enable_management: Client created in management mode, used for administration purposes
            :type enable_management: bool

        """
        self._db = None
        self._vault = None
        self._deploy = None
        self._storage = None
        if config is None:
            config = ConfigManager(use_env=use_env, use_vault=use_vault, config_file=config_file)

        # Create a manager for Vault
        self._create_from_object(config, enable_management)

    def _create_from_object(self, config, enable_management=False):
        """
            Auxiliary method for object constructor

            :param config: Configuration manager object
            :type config: tesla_ce.lib.ConfigManager
            :param enable_management: Client created in management mode, used for administration purposes
            :type enable_management: bool

        """
        self._config = config
        self.management_mode = enable_management
        self._config.config.set('VAULT_MANAGEMENT', enable_management)

        # Only connect managers when running in production mode
        if self._config.config.get('TESLA_MODE') == 'production':
            # Create a manager for Vault
            self._vault = VaultManager(self.config)

            # Create a manager for the database
            self._db = DatabaseManager(self.config)

            # Create a manager for storage
            self._storage = StorageManager(self.config)

        # Create the deployment manager
        self._deploy = DeploymentManager(self)

    @property
    def vault(self):
        """
            Access to the HashiCorp Vault manager
            :return: Instance of the Vault manager
        """
        if self._vault is None:
            self._vault = VaultManager(self.config)
        return self._vault

    @property
    def database(self):
        """
            Access to the database manager
            :return: Instance of the database manager
        """
        if self._db is None:
            self._db = DatabaseManager(self.config)
        return self._db

    @property
    def config(self):
        """
            Access to the configuration manager
            :return: Instance of the configuration manager
        """
        return self._config

    @property
    def deploy(self):
        """
            Access to the deployment manager
            :return: Instance of the deployment manager
        """
        return self._deploy

    @property
    def storage(self):
        """
            Access to the storage manager
            :return: Instance of the storage manager
        """
        return self._storage

    @classmethod
    @tesla_mode_required('config')
    def generate_configuration(cls, output_file, domain, deploy_external_services=False, deploy_moodle=False):
        """
            Generates a configuration file with all the options

            :param output_file: File where configuration will be stored
            :type output_file: str
            :param domain: Base domain that will be used for TeSLA
            :type domain: str
            :param deploy_external_services: If enabled, external services configuration will be generated
            :type deploy_external_services: bool
            :param deploy_moodle: If enabled, moodle configuration will be generated
            :type deploy_moodle: bool
        """
        # Create an empty configuration object
        conf_manager = ConfigManager(load_config=False)

        # Load configuration from environment variables and secrets
        conf_manager.load_env()

        # Load from current file
        if os.path.exists(output_file):
            conf_manager.load_file(output_file)

        # Use this as base configuration
        config = conf_manager.config

        # Store domain information
        config.set('TESLA_DOMAIN', domain)

        # Generate a mail for admin
        config.set('TESLA_ADMIN_MAIL', 'admin@{}'.format(domain))

        # Generate DJango configuration
        config.set('DJANGO_SECRET_KEY', uuid.uuid4().__str__())
        config.set('DJANGO_ALLOWED_HOSTS', domain)

        # If external services are included, generate credentials and configuration
        if deploy_external_services:
            # Enable the services flag
            config.set('DEPLOYMENT_SERVICES', True)

            # Generate TeSLA Database configuration
            config.set('DB_HOST', 'database')
            config.set('DB_NAME', 'tesla')
            config.set('DB_USER', 'tesla')
            config.set('DB_ROOT_PASSWORD', uuid.uuid4().__str__())
            config.set('DB_PASSWORD', uuid.uuid4().__str__())

            # Generate Vault configuration
            config.set('VAULT_URL', 'https://vault.{}'.format(domain))
            config.set('VAULT_DB_HOST', 'database')
            config.set('VAULT_DB_NAME', 'vault')
            config.set('VAULT_DB_USER', 'vault')
            config.set('VAULT_DB_PASSWORD', uuid.uuid4().__str__())

            # Generate Redis configuration
            config.set('REDIS_HOST', 'redis')
            config.set('REDIS_PASSWORD', uuid.uuid4().__str__())

            # Generate MinIO configuration
            config.set('STORAGE_URL', 'https://storage.{}'.format(domain))
            config.set('STORAGE_ACCESS_KEY', uuid.uuid4().__str__())
            config.set('STORAGE_SECRET_KEY', uuid.uuid4().__str__())

            # Generate RabbitMQ configuration
            config.set('RABBITMQ_ERLANG_COOKIE', uuid.uuid4().__str__())
            config.set('RABBITMQ_ADMIN_USER', uuid.uuid4().__str__())
            config.set('RABBITMQ_ADMIN_PASSWORD', uuid.uuid4().__str__())

            # Generate Celery configuration
            config.set('CELERY_BROKER_HOST', 'rabbitmq')
            config.set('CELERY_BROKER_PORT', config.get('RABBITMQ_PORT'))
            config.set('CELERY_BROKER_USER', config.get('RABBITMQ_ADMIN_USER'))
            config.set('CELERY_BROKER_PASSWORD', config.get('RABBITMQ_ADMIN_PASSWORD'))

        # If Moodle is included, generate credentials and configuration
        if deploy_moodle:
            # Enable the moodle flag
            config.set('MOODLE_DEPLOY', True)

            # Generate Moodle Database configuration
            config.set('MOODLE_DB_HOST', config.get('DB_HOST'))
            config.set('MOODLE_DB_PASSWORD', uuid.uuid4().__str__())

            # Generate Moodle administrator password
            config.set('MOODLE_ADMIN_PASSWORD', uuid.uuid4().__str__())

        # Write the configuration file to disk
        with open(output_file, 'w') as out_fh:
            config.write(out_fh)

    def check_configuration(self):
        """
            Check the current configuration

            :return: Report of the configuration check
            :rtype: dict
        """
        storage_status = self.storage.get_status()
        storage_report = {
            'valid': storage_status['status'] == 1,
            'errors': storage_status['info']['errors'],
        }
        report = {
            'tesla': check_tesla_configuration(self.config.config),
            'database': check_db_configuration(self.config.config),
            'vault': check_vault_configuration(self.config.config),
            'storage': storage_report,
        }
        valid = True
        for service in report:
            if not report[service]['valid']:
                valid = False
                break
        report['valid'] = valid

        return report

    @tesla_mode_required('config')
    def initialize(self):
        """
            Initialize the TeSLA CE instance

            :exception TeslaAuthException: If management mode is not enabled
        """
        # Check management mode
        if not self.management_mode:
            raise TeslaAuthException('This operation requires enabled management mode')

        # Check configuration
        report = self.check_configuration()
        if not report['valid']:
            raise TeslaConfigException('Invalid configuration')

        # Initialize Vault
        self.vault.initialize()

        # Initialize the database
        self.database.initialize()

        # Initialize the storage
        self.storage.initialize()

    def get_learner_token_pair(self, learner, scope, filters=None, ttl=15, max_ttl=120):
        """
            Get a JWT token pair for a given learner

            :param leraner: Learner object
            :type learner: tesla_ce.models.Learner
            :param scope: List of tasks allowed for this token
            :type scope: list
            :param filters: Dictionary with the filters applied to the scope
            :type filters: dict
            :param ttl: Time to live for the token in minutes
            :type ttl: int
            :param max_ttl: Maximum time this token can be refreshed
            :type max_ttl: int

            :return: JWT token pair for this learner
            :rtype: dict
        """
        data = {
            'sub': learner.learner_id.__str__(),
            'type': 'learner',
            'pk': learner.id,
            'scope': scope,
            'filters': filters
        }

        return self.vault.create_token_pair(data, ttl=ttl, max_ttl=max_ttl, key='learners')

    def refresh_token(self, access_token, refresh_token):
        """
            Refresh given token.

            :param access_token: The JWT token to be refreshed
            :type access_token: str
            :param refresh_token: The JWT refresh token authenticating this operation
            :type refresh_token: str
            :return: JWT token
            :rtype: str
        """
        try:
            return self.vault.refresh_token(access_token, refresh_token)
        except TeslaVaultException:
            raise TeslaAuthException('Token refresh not authorized')

    def get_user_token_pair(self, user, scope, filters=None, ttl=15, max_ttl=120):
        """
            Get a JWT token pair for a given user

            :param user: Institution user object
            :type user: tesla_ce.models.InstitutionUser
            :param scope: List of tasks allowed for this token
            :type scope: list
            :param filters: Dictionary with the filters applied to the scope
            :type filters: dict
            :param ttl: Time to live for the token in minutes
            :type ttl: int
            :param max_ttl: Maximum time this token can be refreshed
            :type max_ttl: int

            :return: JWT token pair for this user
            :rtype dict
        """
        data = {
            'sub': user.uid,
            'type': 'user',
            'pk': user.id,
            'scope': scope,
            'filters': filters
        }

        return self.vault.create_token_pair(data, ttl=ttl, max_ttl=max_ttl, key='users')

    def get_admin_token_pair(self, user, scope, filters=None, ttl=15, max_ttl=120):
        """
            Get a JWT token pair for a given user

            :param user: User object
            :type user: tesla_ce.models.User
            :param scope: List of tasks allowed for this token
            :type scope: list
            :param filters: Dictionary with the filters applied to the scope
            :type filters: dict
            :param ttl: Time to live for the token in minutes
            :type ttl: int
            :param max_ttl: Maximum time this token can be refreshed
            :type max_ttl: int

            :return: JWT token pair for this user
            :rtype dict
        """
        if not user.is_staff:
            raise TeslaAuthException('User is not admin')
        data = {
            'sub': user.username,
            'type': 'admin',
            'pk': user.id,
            'scope': scope,
            'filters': filters
        }

        return self.vault.create_token_pair(data, ttl=ttl, max_ttl=max_ttl, key='users')

    def get_instructor_token(self, mail, scope, filters=None, ttl=60):
        """
            Get a JWT token for an instructor

            :param mail: Instructor email
            :type mail: str
            :param scope: List of tasks allowed for this token
            :type scope: list
            :param filters: Dictionary with the filters applied to the scope
            :type filters: dict
            :param ttl: Time to live for the token in minutes
            :type ttl: int

            :return: JWT token for this instructor
        """
        data = {
            'sub': mail,
        }

        return self.vault.create_token(data, ttl=ttl, key='instructors')

    def get_module_token(self, scope, module_id=None, filters=None, ttl=60, data=None):
        """
            Get a JWT token for a module

            :param module_id: Module unique identifier
            :type module_id: str
            :param scope: List of tasks allowed for this token
            :type scope: list
            :param filters: Dictionary with the filters applied to the scope
            :type filters: dict
            :param ttl: Time to live for the token in minutes
            :type ttl: int
            :param data: Information to add on the payload of the token
            :type data: dict
            :return: JWT token for this module
        """
        if module_id is None:
            module_info = self.config.get_module()
            module_id = module_info['module']
        if data is None:
            data = {}
        # Ensure required module information
        if module_id.startswith('provider_') or module_id.startswith('vle_'):
            id_parts = module_id.split('_')
            if 'type' not in data:
                data['type'] = id_parts[0]
            if 'pk' not in data:
                data['pk'] = int(id_parts[1])
        data['sub'] = module_id
        data['scope'] = scope
        data['filters'] = filters

        return self.vault.create_token(data, ttl=ttl, key='module_{}'.format(module_id))

    @cache_memoize(60 * 15)
    def validate_token(self, token):
        """
            Validate a token and return its data

            :param token: The token to be validated
            :type token: str
            :return: Token data
            :rtype: dict
            :exception TeslaAuthException: In case the token is invalid or expired
        """
        try:
            validation = self.vault.validate_token(token)
        except TeslaVaultException:
            raise TeslaAuthException('Invalid JWT token')

        if validation['valid']:
            return validation['payload']

        raise TeslaAuthException('Invalid JWT token')

    def verify_user(self, email, password):
        """
            Validate user credentials

            :param email: The user email
            :type email: str
            :param password: The user password
            :type password: str
            :return: User entity data
            :rtype: dict
            :exception TeslaAuthException: In case the credentials are not valid
        """
        try:
            user_info = self.vault.verify_user_password(email, password)
            user = models.User.objects.get(email=user_info['email'])
        except TeslaVaultException:
            raise TeslaAuthException('Invalid user credentials')
        except models.User.DoesNotExist:
            raise TeslaAuthException('Invalid user credentials')

        if user.is_staff:
            token_pair = self.get_admin_token_pair(user=user,
                                                   scope=['/api/v2/admin/*', '/api/v2/institution/*'],
                                                   max_ttl=60)
        elif hasattr(user, 'institutionuser'):
            inst_user = user.institutionuser
            scopes = ['/api/v2/institution/{}/user/{}/*'.format(
                inst_user.institution_id, inst_user.id
            )]
            if inst_user.inst_admin:
                scopes.append('/api/v2/institution/{}/*'.format(
                    inst_user.institution_id
                ))
            if inst_user.send_admin:
                scopes.append('/api/v2/institution/{}/send/*'.format(
                    inst_user.institution_id
                ))
            if inst_user.legal_admin:
                scopes.append('/api/v2/institution/{}/ic/*'.format(
                    inst_user.institution_id
                ))
            if 'LEARNER' in models.user.get_institution_roles(user):
                learner_id = str(inst_user.learner.learner_id)
                scopes.append('/lapi/v1/enrolment/{}/{}/'.format(
                    inst_user.institution_id,
                    learner_id
                ))
                scopes.append('/lapi/v1/status/{}/{}/'.format(
                    inst_user.institution_id,
                    learner_id
                ))
                scopes.append('/lapi/v1/alert/{}/{}/'.format(
                    inst_user.institution_id,
                    learner_id
                ))
                scopes.append('/api/v2/institution/{}/learner/{}/*'.format(
                    inst_user.institution_id, inst_user.id
                ))
            token_pair = self.get_user_token_pair(user=inst_user, scope=scopes, ttl=15, max_ttl=24*60)
        else:
            raise TeslaAuthException('Invalid user. Missing Institution or administration rights.')

        return token_pair

    def get_system_status(self):
        """
            Get the system status

            :return: Object with overall status and specific status for each module
            :rtype: dict
        """
        report = {
            'db': self.database.get_status(),
            'vault': self.vault.get_status(),
            'config': self.config.get_status(),
            'storage': self.storage.get_status(),
        }

        # Compute summary
        status = 1
        errors = 0
        warnings = 0
        for module in report:
            if report[module]['status'] == 0:
                status = 0
            errors += report[module]['errors']
            warnings += report[module]['warnings']

        return {
            'status': status,
            'warnings': warnings,
            'errors': errors,
            'info': {},
            'report': report
        }

    def export_deployment_scripts(self, output, mode=None):
        """
            Generates the files required to deploy TeSLA CE on a certain docker orchestrator

            :param output: Path to the folder where the scripts will be written
            :type output: str
            :param mode: The deployment mode (orchestrator) to be used
            :type mode: str
        """
        # Get the orchestrator
        if mode is None:
            mode = self.config.config.get('DEPLOYMENT_ORCHESTRATOR')

        files = self.deploy.get_deployment_scripts(mode)

        # Write files to disk
        if not os.path.exists(output):
            os.makedirs(output)
        for file in files:
            out_file = os.path.join(output, file)
            if not os.path.exists(os.path.dirname(out_file)):
                os.makedirs(os.path.dirname(out_file))
            with open(out_file, 'w') as out_fh:
                out_fh.write(files[file])

    def export_services_scripts(self, output, mode=None):
        """
            Generates the files required to deploy TeSLA CE required services on a certain docker orchestrator

            :param output: Path to the folder where the scripts will be written
            :type output: str
            :param mode: The deployment mode (orchestrator) to be used
            :type mode: str
        """
        # Get the orchestrator
        if mode is None:
            mode = self.config.config.get('DEPLOYMENT_ORCHESTRATOR')

        files = self.deploy.get_services_deployment_scripts(mode)

        # Write files to disk
        if not os.path.exists(output):
            os.makedirs(output)
        for file in files:
            out_file = os.path.join(output, file)
            if not os.path.exists(os.path.dirname(out_file)):
                os.makedirs(os.path.dirname(out_file))
            with open(out_file, 'w') as out_fh:
                out_fh.write(files[file])

    def export_vle_scripts(self, type, client_id, output, mode=None):
        """
            Generates the files required to deploy TeSLA CE VLE on a certain docker orchestrator

            :param type: Type of VLE. Accepted values are: 'moodle'
            :type type: str
            :param client_id: LTI 1.3 Client ID for this VLE
            :type client_id: str
            :param output: Path to the folder where the scripts will be written
            :type output: str
            :param mode: The deployment mode (orchestrator) to be used
            :type mode: str
        """
        # Get the orchestrator
        if mode is None:
            mode = self.config.config.get('DEPLOYMENT_ORCHESTRATOR')

        # Check the VLE type
        type_id = models.vle.get_type_id(type)
        if type_id is None:
            raise TeslaConfigException('Invalid VLE type')

        name = self.config.config.get('{}_NAME'.format(type.upper()))
        deploy = self.config.config.get('{}_DEPLOY'.format(type.upper()))
        if not deploy:
            raise TeslaConfigException('This VLE is not configured to be deployed')

        # Create the database and the user
        self.database.create_database(self.config.config.get('MOODLE_DB_NAME'))
        self.database.create_user(self.config.config.get('MOODLE_DB_NAME'),
                                  self.config.config.get('MOODLE_DB_USER'),
                                  self.config.config.get('MOODLE_DB_PASSWORD'))

        # Find VLE
        try:
            vle = models.VLE.objects.get(name=name)
        except models.VLE.DoesNotExist:
            institution = models.Institution.objects.first()
            vle = models.VLE.objects.create(name=name, type=type_id, url="", institution=institution,
                                            client_id=client_id)
            vle.save()
        files = self.deploy.get_vle_deployment_scripts(vle, mode)

        # Write files to disk
        if not os.path.exists(output):
            os.makedirs(output)
        for file in files:
            out_file = os.path.join(output, file)
            if not os.path.exists(os.path.dirname(out_file)):
                os.makedirs(os.path.dirname(out_file))
            with open(out_file, 'w') as out_fh:
                out_fh.write(files[file])

    def create_assessment_session(self, activity, learner, locale=None, max_ttl=120, redirect_reject_url=None,
                                  reject_message=None, options=None):
        """
            Create a new assessment session for a given activity and learner

            :param activity: The activity for the assessment session
            :type activity: tesla_ce.models.Activity
            :param learner: The learner for the assessment session
            :type learner: tesla_ce.models.Learner
            :param locale: Fix the locale to use for this assessment
            :type locale: str
            :param max_ttl: The maximum validity of provided tokens in minutes
            :type max_ttl: int
            :param redirect_reject_url: Redirect URL where learner is redirected if ethical warning is rejected
            :type redirect_reject_url: str
            :param reject_message: Message provided to learner in case ethical warning is rejected
            :type reject_message: str
            :param options: Options that can be passed to the data object to modify some default properties
            :return: Created session
            :rtype: tesla_ce.models.AssessmentSession
        """
        # Check learner informed consent status
        if not learner.institution.external_ic:
            if learner.consent_accepted is None or learner.consent_rejected is not None:
                raise TeslaMissingICException()

            if not learner.ic_status.startswith('VALID'):
                raise TeslaInvalidICException()

        # Get the list of instruments and enrolment values
        enrolment_obj = {
            'missing_enrolments': False,
            'instruments': {}
        }
        if activity.enabled:
            # Check enrolment status for this learner and activity
            enrolment_obj = learner.missing_enrolments(activity.id)
            if enrolment_obj['missing_enrolments']:
                raise TeslaMissingEnrolmentException(enrolment_obj)

        # Set the list of sensors
        activity_instruments = activity.get_learner_instruments(learner)
        sensors = {}
        instruments = []
        for instrument in activity_instruments:
            instruments.append(instrument.instrument_id)
            inst_sensor = instrument.get_sensors()
            for sensor in inst_sensor:
                if sensor not in sensors:
                    sensors[sensor] = []
                sensors[sensor].append(instrument.instrument_id)

        # Initialize a new Assessment Session
        session = models.AssessmentSession.objects.create(activity=activity, learner=learner)

        # Add the learner to the course
        activity.course.learners.add(learner)
        activity.course.save()

        # Get the learner token
        token = self.get_learner_token_pair(learner,
                                            [
                                                '/lapi/v1/verification/{}/{}/'.format(
                                                    learner.institution_id, learner.learner_id),
                                                '/lapi/v1/status/{}/{}/'.format(
                                                    learner.institution_id, learner.learner_id),
                                                '/lapi/v1/alert/{}/{}/'.format(
                                                    learner.institution_id, learner.learner_id)
                                            ], ttl=15, max_ttl=max_ttl,
                                            filters={
                                                "session_id": session.id,
                                                "mode": "verification"
                                            })

        # Create a Dashboard launcher
        launcher = self.get_launcher_token('dashboard', learner.institutionuser, session)

        # Compute the base path for assets
        base_url = '{}'.format(static('web-plugin/web-plugin.js').split('web-plugin.js')[0])

        # Set default options
        if options is None:
            options = {}
        if 'floating_menu_initial_pos' not in options:
            options['floating_menu_initial_pos'] = 'top-right'

        # Update the information
        session_connector_data = {
            'mode': 'verification',
            'learner': {
                'id': learner.id,
                'learner_id': learner.learner_id.__str__(),
                'first_name': learner.first_name,
                'last_name': learner.last_name,
                'picture': None,
                'institution_id': learner.institution_id,
            },
            'session_id': session.id,
            'activity': {
                'enabled': activity.enabled,
                'course': {
                    'id': activity.course.id,
                    'code': activity.course.code,
                    'description': activity.course.description,
                    'start': activity.course.start.isoformat() if activity.course.start is not None else None,
                    'end': activity.course.end.isoformat() if activity.course.end is not None else None,
                },
                'id': activity.id,
                'name': activity.name,
                'description': activity.description,
                'start': activity.start.isoformat() if activity.start is not None else None,
                'end': activity.end.isoformat() if activity.end is not None else None,
                'allow_reject_capture': False,
                'redirect_reject_url': redirect_reject_url,
                'rejected_message': reject_message,
                'vle_id': activity.vle_id,
            },
            'accessibility': {
                'high_contrast': False,
                'big_fonts': False,
                'text_to_speech': False
            },
            'instruments': instruments,
            'sensors': sensors,
            'token': token,
            'enrolment': enrolment_obj,
            'logo_url': static('img/android-chrome-192x192.png'),
            'api_url': settings.API_URL,
            'dashboard_url': settings.DASHBOARD_URL,
            'launcher': launcher,
            'base_url': base_url,
            'locale': locale,
            'options': options
        }

        # Store the session data
        session_data = models.AssessmentSessionData.objects.create(session=session)
        session_connector_data_path = os.path.join(str(learner.institution.id),
                                                   str(learner.learner_id),
                                                   'sessions',
                                                   str(session.activity_id),
                                                   str(session.id), 'data.json'
                                                   )
        session_data.data.save(session_connector_data_path,
                               ContentFile(simplejson.dumps(session_connector_data).encode('utf-8')))

        # Generate the connector
        connector_data = {
            'data_url': session_data.data.url,
            'modules': [
                static('web-plugin/web-plugin.js'),
            ],
            'css': [
                static('web-plugin/styles.css'),
            ],
            'active': activity.enabled and len(instruments) > 0
        }
        session_connector = render_to_string('web-plugin/connector.js',
                                             {'session_data': simplejson.dumps(connector_data)})

        # Create the data path
        session_connector_path = os.path.join(str(learner.institution.id),
                                              str(learner.learner_id),
                                              'sessions',
                                              str(session.activity_id),
                                              str(session.id), 'connector.js'
                                    )
        # Write assessment info file
        session_data.connector.save(session_connector_path, ContentFile(session_connector.encode('utf-8')))
        session_data.save()

        return session

    def register_vle(self, type, name, url=None, institution_acronym=None, client_id=None, force_update=False):
        """
            Register a new VLE to the system

            :param type: Type of VLE. Accepted values are: 'moodle'
            :type type: str
            :param name: Unique name for the VLE
            :type name: str
            :param url: URL of the VLE
            :type url: str
            :param institution_acronym: Acronym of the institution whit VLE belongs to
            :type institution_acronym: str
            :param client_id: LTI 1.3 Client ID for this VLE
            :type client_id: str
            :param force_update: If the VLE already exists, update the data
            :type force_update: bool
            :return: Configuration data for the new VLE
            :rtype: dict
        """
        # Check the institution
        if institution_acronym is not None:
            try:
                institution = models.Institution.objects.get(acronym=institution_acronym)
            except models.Institution.DoesNotExist:
                raise TeslaConfigException('Invalid institution acronym {}'.format(institution_acronym))
        else:
            if models.Institution.objects.count() > 1:
                raise TeslaConfigException('Multiple registered institutions. Institution acronym is required')
            institution = models.Institution.objects.first()

        # Check the VLE type
        type_id = models.vle.get_type_id(type)
        if type_id is None:
            raise TeslaConfigException('Invalid VLE type')

        # Create the register on the database
        try:
            vle = models.VLE.objects.get(name=name)
            if vle is not None and not force_update:
                raise TeslaConfigException(
                    'A VLE with this name already exist. User force_update to update current vle information'
                )
        except models.VLE.DoesNotExist:
            vle = models.VLE.objects.create(name=name, type=type_id, url=url, institution=institution,
                                            client_id=client_id)
            vle.save()

        # Update LTI data
        if vle.lti is None:
            lti = {}
        else:
            lti = json.loads(vle.lti)

        # LTI 1.1 data
        if 'consumer_key' not in lti or lti['consumer_key'] is None or len(lti['consumer_key']) == 0:
            lti['consumer_key'] = uuid.uuid4().__str__()
        if 'consumer_secret' not in lti or lti['consumer_secret'] is None or len(lti['consumer_secret']) == 0:
            lti['consumer_secret'] = uuid.uuid4().__str__()

        # LTI 1.3 data
        # Update URLs for known VLEs
        if url is not None and type.upper() == 'MOODLE':
            lti['auth_login_url'] = '{}/mod/lti/auth.php'.format(url)
            lti['auth_token_url'] = '{}/mod/lti/token.php'.format(url)
            lti['key_set_url'] = '{}/mod/lti/certs.php'.format(url)

        # Update the VLE object
        vle.lti = json.dumps(lti)
        vle.save()

        # Register the vle
        vle_info = self.vault.register_vle(vle)
        vle_info.update(lti)

        # Add API URL
        vle_info['api_url'] = 'https://{}'.format(self.config.config.get('TESLA_DOMAIN'))

        return vle_info

    def register_provider(self, provider_info, force_update=False):
        """
            Register a new Provider to the system

            :param provider_info: Provider information. Should be provided by the provider
            :type provider_info: dict
            :param force_update: If the Provider already exists, update the data
            :type force_update: bool
            :return: Configuration data for the new Provider
            :rtype: dict
        """
        try:
            provider = models.Provider.objects.get(instrument_id=provider_info['instrument'],
                                                   acronym=provider_info['acronym'])
            provider.image = provider_info['image']
            provider.name = provider_info['name']
            provider.version = provider_info['version']
            provider.queue = provider_info['queue']
            provider.url = provider_info.get('url'),
            provider.description = provider_info['description']
            provider.has_service = provider_info.get('has_service', False)
            provider.service_port = provider_info.get('service_port')
            provider.options_schema = provider_info.get('options_schema')
            provider.allow_validation = provider_info.get('allow_validation', False)
            provider.alert_below = provider_info.get('alert_below')
            provider.warning_below = provider_info.get('warning_below')
            provider.inverted_polarity = provider_info.get('inverted_polarity', False)
            provider.save()
        except models.Provider.DoesNotExist:
            provider = models.Provider.objects.create(
                instrument_id=provider_info['instrument'],
                acronym=provider_info['acronym'],
                image=provider_info['image'],
                name=provider_info['name'],
                version = provider_info['version'],
                queue = provider_info['queue'],
                url=provider_info.get('url'),
                description=provider_info['description'],
                has_service=provider_info.get('has_service', False),
                service_port=provider_info.get('service_port'),
                options_schema=provider_info.get('options_schema'),
                allow_validation=provider_info.get('allow_validation', False),
                alert_below=provider_info.get('alert_below'),
                warning_below=provider_info.get('warning_below'),
                inverted_polarity=provider_info.get('inverted_polarity', False)
            )
        # Register the provider module
        prov_info = self.vault.register_provider(provider)
        return prov_info

    def get_launcher_token(self, target, user, session=None, target_url=None, ttl=None):
        """
            Create a launcher token

            :param target:
            :param user:
            :param session:
            :param target_url:
            :param ttl:
            :return: Launcher credentials
            :rtype: dict
        """
        target_id = None
        token_pair = None
        if target.upper() == 'DASHBOARD':
            target_id = 0
            if ttl is None:
                ttl = 24 * 60
            token_pair = self.get_user_token_pair(user=user, scope=['/api/v2/institution/{}/user/{}/*'.format(
                user.institution_id, user.id
            )], max_ttl=ttl)
        elif target.upper() == 'LAPI':
            target_id = 1
            if ttl is None:
                ttl = 2 * 60
            try:
                learner = user.learner
            except user.learner.RelatedObjectDoesNotExist:
                raise TeslaAuthException('Provided user is not a learner')
            token_pair = self.get_learner_token_pair(learner=learner, scope=['/lapi/*'], max_ttl=ttl)

        if target_id is None or token_pair is None:
            raise TeslaConfigException('Invalid target type')

        expiration = timezone.now() + timezone.timedelta(minutes=ttl)

        launcher =  models.Launcher.objects.create(user=user, token=uuid.uuid4(), session=session, target_url=target_url,
                                                   token_pair=token_pair, expires_at=expiration)

        return {
            'id': launcher.id,
            'token': launcher.token.__str__()
        }

    def get_registered_providers(self, repository="tesla-ce/core", version="main"):
        """
            Get the list of registered providers

            :param repository: The repository in GitHub
            :type repository: str
            :param version: The branch on the repository
            :type version: str
            :return: List of registered providers
            :rtype: list
        """
        return self.deploy.get_registered_providers(repository, version)

    def export_provider_scripts(self, acronym, output, credentials=None, mode=None):
        """
            Generates the files required to deploy TeSLA CE VLE on a certain docker orchestrator

            :param acronym: Provider acronym
            :type type: str
            :param output: Path to the folder where the scripts will be written
            :type output: str
            :param credentials: List of credentials required by this provider
            :type credentials: list
            :param mode: The deployment mode (orchestrator) to be used
            :type mode: str
        """
        # Get the orchestrator
        if mode is None:
            mode = self.config.config.get('DEPLOYMENT_ORCHESTRATOR')

        # Find Provider
        try:
            provider = models.Provider.objects.get(acronym=acronym)
        except models.Provider.DoesNotExist:
            raise TeslaConfigException('Invalid provider acronym')
        files = self.deploy.get_provider_deployment_scripts(provider, mode, credentials=credentials)

        # Write files to disk
        if not os.path.exists(output):
            os.makedirs(output)
        for file in files:
            out_file = os.path.join(output, file)
            if not os.path.exists(os.path.dirname(out_file)):
                os.makedirs(os.path.dirname(out_file))
            with open(out_file, 'w') as out_fh:
                out_fh.write(files[file])



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
from django.conf import settings

from tesla_ce.models import AuthenticatedUser
from tesla_ce.models import AuthenticatedModule
from tesla_ce.models import Learner
from tesla_ce.models import InstitutionUser


def is_authenticated(request, institution_id, learner_id, data):

    if data['learner_id'] != str(learner_id):
        return False
    try:
        if not request.user.is_authenticated:
            return False
        if isinstance(request.user, AuthenticatedUser):
            if request.user.type == 'learner' and request.user.sub == str(learner_id):
                return True
            return request.user.type == 'user' and request.path in request.user.allowed_scopes
        if isinstance(request.user, Learner):
            return str(request.user.learner_id) == str(learner_id) and request.user.institution_id == institution_id
        if isinstance(request.user, InstitutionUser):
            return str(request.user.learner.learner_id) == str(learner_id) and \
                   request.user.learner.institution_id == institution_id
        if isinstance(request.user, AuthenticatedModule) and request.user.type == 'vle':
            return request.user.module.institution_id == institution_id

    except request.user._meta.model.learner.RelatedObjectDoesNotExist:
        # Current user is not a learner
        return False
    return False



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" InformedConsent views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.views import Response
from rest_framework.views import status
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api.v2.serializers import InstitutionInformedConsentSerializer
from tesla_ce.models import InformedConsent
from tesla_ce.models import Institution
from tesla_ce.models.informed_consent import get_current_ic_version


# pylint: disable=too-many-ancestors
class InstitutionInformedConsentViewSet(viewsets.ModelViewSet, NestedViewSetMixin):
    """
    API endpoint that allows activity to be viewed or edited.
    """
    serializer_class = InstitutionInformedConsentSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    '''
    filterset_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    search_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    '''

    def get_queryset(self):
        queryset = InformedConsent.objects
        if 'parent_lookup_institution_id' in self.kwargs:
            queryset = queryset.filter(
                institution__id=self.kwargs['parent_lookup_institution_id']
            )
        return queryset.all().order_by('id')

    @action(detail=False, methods=['GET'])
    def current(self, request, *args, **kwargs):
        """
            Retrieve the current informed consent
        """
        try:
            institution = Institution.objects.get(id=kwargs['parent_lookup_institution_id'])
        except Institution.DoesNotExist:
            return Response('Invalid institution', status=status.HTTP_404_NOT_FOUND)
        try:
            current_ic = InformedConsent.objects.get(version=get_current_ic_version(institution),
                                                     institution_id=kwargs['parent_lookup_institution_id'])
        except InformedConsent.DoesNotExist:
            return Response('Invalid version', status=status.HTTP_404_NOT_FOUND)

        serializer = self.get_serializer(current_ic)
        return Response(serializer.data)



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Base Webhook client implementation """
import abc
import importlib

from django.utils import timezone

from tesla_ce.models import WebhookClient
from tesla_ce.models import WebhookMessage
from .. import exceptions


class BaseWebhookMessage:
    def __init__(self, client, body, message_id=None, model=None):
        self.body = body
        self.message_id = message_id
        self.client = client
        self.model = model

    @staticmethod
    def _get_class(implementation_class):
        implementation_class_parts = implementation_class.split('.')
        module_name = '.'.join(implementation_class_parts[:-1])
        class_name = implementation_class_parts[-1]

        module = importlib.import_module(module_name)
        try:
            imp_class = getattr(module, class_name)
        except Exception:
            raise exceptions.WebhookImplementationNotFoundException(
                'Invalid implementation class: {}'.format(implementation_class)
            )

        return imp_class

    @staticmethod
    def get_webhook_objects(request):
        # Get the list of client headers
        headers = [header.upper() for header in request.headers]
        clients = WebhookClient.objects.filter(enabled=True, client_header__in=headers).all()
        if len(clients) == 0:
            raise exceptions.WebhookClientNotFoundException('No client found for this request')
        objects = []
        for client in clients:
            message_id = None
            if client.id_header is not None and client.id_header in headers:
                message_id = request.headers[client.id_header]
            imp_class = BaseWebhookMessage._get_class(client.implementation)
            obj = imp_class(client, request.data, message_id)
            obj.authenticate(request)
            objects.append(obj)

        return objects

    @staticmethod
    def get_webhook_object(request):
        """
            Get a client object from request
            :param request: The HTTP request with webhook message
            :return: BaseCle
        """

        # Get the list of clients for current request
        clients = BaseWebhookMessage.get_webhook_objects(request)

        if clients is None:
            return None
        if len(clients) > 1:
            raise exceptions.WebhookMultipleClientsException('Multiple clients for given request')

        return clients[0]

    @abc.abstractmethod
    def _process_webhook(self):
        raise NotImplemented('Specific webhook process is not implemented')

    @abc.abstractmethod
    def _is_authenticated(self, request):
        raise NotImplemented('Specific webhook process is not implemented')

    def process(self):
        self.model.status = 1
        self.model.start_process_at = timezone.now()
        self.save()
        try:
            self._process_webhook()
        except Exception as exc:
            self.update_status(3, exc.__str__())

    def authenticate(self, request):
        if not self._is_authenticated(request):
            raise exceptions.WebhookAuthException('Invalid authentication')

    @staticmethod
    def load(pk, request=None):
        try:
            model = WebhookMessage.objects.get(pk=pk)
        except WebhookMessage.DoesNotExist:
            return None
        imp_class = BaseWebhookMessage._get_class(model.client.implementation)
        obj = imp_class(model.client, model.body, model.message_id, model)
        if request is not None:
            obj.authenticate(request)
        return obj

    def save(self):
        if self.model is None:
            self.model = WebhookMessage.objects.create(
                status=0,
                message_id=self.message_id,
                client=self.client,
                body=self.body
            )
        else:
            self.model.save()

    def get_data(self):
        if self.model is None:
            return None
        return self.model.get_body()

    def update_status(self, status, message):
        self.model.status = status
        self.model.error = message
        self.save()



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Worker actions for Use Case tests """
import mock


def get_task_by_queue(tasks):
    """
        Split a list of tasks on their respective queues
        :param tasks: List of tasks
        :return: Dictionary with tasks organized in queues
    """
    queues = {}
    for task in tasks:
        if task[2]['queue'] not in queues:
            queues[task[2]['queue']] = []
        queues[task[2]['queue']].append(task)
    return queues


def worker_validation_summary(tasks):
    """
    Worker compute validation summary from individual validations
    :param tasks: Pending tasks to be processed
    :return: List of new pending tasks
    """
    # List simulating the enrolment queue
    pending_tasks_enrolment = []

    def enrol_learner_test(*args, **kwargs):
        pending_tasks_enrolment.append(('enrol_learner', args, kwargs))

    # Run validation summary tasks
    with mock.patch('tesla_ce.tasks.requests.enrolment.enrol_learner.apply_async', enrol_learner_test):
        from tesla_ce.tasks.requests.enrolment import create_validation_summary
        for task in tasks:
            assert task[0] == 'create_validation_summary'
            create_validation_summary(*task[1][0])

    return pending_tasks_enrolment


def worker_enrol_learner(tasks):
    """
        Worker distribute enrolment tasks among providers
        :param tasks: List of pending tasks
        :return: List of tasks assigned to each provider
    """
    # List simulating the enrolment queue
    pending_tasks_enrolment = []

    def enrol_learner_test(*args, **kwargs):
        pending_tasks_enrolment.append(('enrol_learner', args, kwargs))

    # Run validation summary tasks
    with mock.patch('tesla_ce.tasks.requests.enrolment.enrol_learner.apply_async', enrol_learner_test):
        from tesla_ce.tasks.requests.enrolment import enrol_learner
        for task in tasks:
            assert task[0] == 'enrol_learner'
            enrol_learner(*task[1][0])

    return pending_tasks_enrolment


def worker_send_notifications():
    """
        Worker process notification tasks
        :return: List of pending provider notification tasks
    """
    # List simulating the notifications queue
    pending_tasks_notification = []

    def provider_notify_test(*args, **kwargs):
        pending_tasks_notification.append(('provider_notify', args, kwargs))

    # Process notifications
    with mock.patch('tesla_ce.tasks.notification.providers.provider_notify.apply_async',
                    provider_notify_test):
        from tesla_ce.tasks.notification.providers import send_provider_notifications
        send_provider_notifications()

    return pending_tasks_notification


def worker_create_reports(tasks):
    """
        Worker process verification results to create the reports
        :param tasks: List of pending reporting tasks
    """
    # List simulating the reporting queue
    pending_tasks_reporting = []

    # List simulating the reporting queue
    pending_tasks_reporting2 = []

    def update_learner_activity_instrument_report_test(*args, **kwargs):
        pending_tasks_reporting.append(('update_learner_activity_instrument_report', args, kwargs))

    def update_learner_activity_report_test(*args, **kwargs):
        pending_tasks_reporting2.append(('update_learner_activity_report', args, kwargs))

    for task in tasks:
        assert task[0] == 'create_verification_summary'
        with mock.patch(
                'tesla_ce.tasks.reports.results.update_learner_activity_instrument_report.apply_async',
                update_learner_activity_instrument_report_test):
            from tesla_ce.tasks.requests.verification import create_verification_summary
            create_verification_summary(*task[1][0])

    # Perform reporting tasks
    for task in pending_tasks_reporting:
        assert task[0] == 'update_learner_activity_instrument_report'
        with mock.patch(
                'tesla_ce.tasks.reports.results.update_learner_activity_report.apply_async',
                update_learner_activity_report_test):
            from tesla_ce.tasks.reports.results import update_learner_activity_instrument_report
            update_learner_activity_instrument_report(*task[1][0])

    # Perform reporting tasks
    for task in pending_tasks_reporting2:
        assert task[0] == 'update_learner_activity_report'
        from tesla_ce.tasks.reports.results import update_learner_activity_report
        from celery.exceptions import Reject
        try:
            update_learner_activity_report(*task[1][0])
        except Reject:
            # Raised when no new data is available
            pass



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Message model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel


class Message(BaseModel):
    """ Message model. """
    code = models.CharField(max_length=50, null=False, blank=False, primary_key=True, help_text=_("Message code"))
    datatype = models.CharField(max_length=50, null=False, blank=False, default='html',
                                help_text=_("Type of the message"))
    description = models.TextField(null=True, blank=False, default=None,
                                   help_text=_("Description of the message for translation"))
    meaning = models.TextField(null=True, blank=False, default=None,
                               help_text=_("Meaning of the message for translation"))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<Message(code='%s')>" % (
            self.code
        )

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.code = self.code.upper()
        super().save(force_insert, force_update, using, update_fields)


class MessageLocale(BaseModel):
    """ Message model. """
    code = models.ForeignKey(Message, null=False, blank=False, on_delete=models.CASCADE, help_text=_("Message code"))
    locale = models.CharField(max_length=10, null=False, blank=False, help_text=_("Locale"))
    message = models.TextField(null=False, blank=False, help_text=_('Message'))
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('code', 'locale'),)

    def __repr__(self):
        return "<MessageLocale(code='%s', locale='%r', message='%r')>" % (
            self.code, self.locale, self.message
        )

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.locale = self.locale.lower()
        super().save(force_insert, force_update, using, update_fields)



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Instrument model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel


class Instrument(BaseModel):
    """ Instrument model. """
    name = models.CharField(max_length=250, null=False, blank=False,
                            help_text=_("Instrument name."))

    acronym = models.CharField(max_length=30, null=False, blank=False, unique=True,
                               help_text=_("Instrument Acronym."))

    queue = models.TextField(null=False, blank=False, unique=True,
                             help_text=_("Queue this instrument listens to."))

    enabled = models.BooleanField(null=False, blank=False, default=False,
                                  help_text=_("The instrument is enabled"))

    requires_enrolment = models.BooleanField(null=False, blank=False, default=False,
                                             help_text=_("Whether this instrument requires enrolment"))

    description = models.TextField(null=True, blank=True, default=None,
                                   help_text=_("Description of the instrument."))

    identity = models.BooleanField(null=False, blank=False, default=False,
                                   help_text=_("This instrument contributes to the learner identity verification"))
    originality = models.BooleanField(null=False, blank=False, default=False,
                                      help_text=_(
                                          "This instrument contributes to the assessment originality verification"))
    authorship = models.BooleanField(null=False, blank=False, default=False,
                                     help_text=_(
                                         "This instrument contributes to the assessment authorship verification"))
    integrity = models.BooleanField(null=False, blank=False, default=False,
                                    help_text=_(
                                         "This instrument contributes to the assessment integrity verification"))
    options_schema = models.TextField(null=True, blank=False, default=None,
                                      help_text=_("Schema for instrument options"))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<Instrument(id='%d', name='%s', acronym='%s', queue='%s', enabled='%r')>" % (
            self.id, self.name, self.acronym, self.queue, self.enabled)

    @property
    def is_active(self):
        """
            Whether the instrument is active. This means that is enabled and have enabled providers.

            :return: True if the instrument is active or false otherwise
            :rtype: bool
        """
        return self.enabled and self.provider_set.filter(enabled=True).count() > 0



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Enrolment sample validation model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .enrolment_sample import EnrolmentSample
from .learner import get_learner_enrolment
from .message import Message
from .provider import Provider

VALIDATION_STATUS = (
    (0, _('Validating')),
    (1, _('Valid')),
    (2, _('Error')),
    (3, _('Timeout')),
    (4, _('Waiting External Service')),
)


class EnrolmentSampleValidation(BaseModel):
    """ Enrolment Sample Validation model. """
    sample = models.ForeignKey(EnrolmentSample, null=False, on_delete=models.CASCADE,
                               help_text=_("Related Sample"))

    provider = models.ForeignKey(Provider, null=False, on_delete=models.CASCADE,
                                 help_text=_("Provider that performed the validation of the sample"))

    status = models.IntegerField(choices=VALIDATION_STATUS, null=False,
                                 default=0,
                                 help_text=_('Validation status for this sample'))

    info = models.FileField(max_length=250, null=False, blank=False,
                            help_text=_("Validation information path on storage."))

    contribution = models.DecimalField(null=True, decimal_places=2, max_digits=5, default=None,
                                       help_text=_('Estimated contribution of this sample to the enrolment model'))

    error_message = models.TextField(null=True, blank=None, default=None,
                                     help_text=_("Error message when status is error"))

    message_code = models.ForeignKey(Message, null=True, default=None,
                                     on_delete=models.SET_NULL,
                                     help_text=_("Related message code"))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('sample', 'provider'),)

    def __repr__(self):
        return "<EnrolmentSampleValidation(sample='%s', provider='%s', status='%r')>" % (
            self.sample_id, self.provider_id, self.get_status_display())

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        super().save(force_insert, force_update, using, update_fields)
        # Invalidate cached value
        get_learner_enrolment.invalidate(self.sample.learner_id)



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Activity Attachment serialize module """
from rest_framework import serializers
from tesla_ce.apps.lapi.serializers import VerificationSampleSerializer


class VLECourseActivityAttachmentSerializer(VerificationSampleSerializer):
    """Activity attachment serialize model module."""

    learner_id = serializers.UUIDField(required=False, allow_null=True, default=None)
    course_id = serializers.IntegerField(required=False, allow_null=True, default=None)
    activity_id = serializers.IntegerField(required=False, allow_null=True, default=None)
    close_session = serializers.BooleanField(required=False, allow_null=True, default=True, write_only=True)
    close_at = serializers.DateTimeField(required=False, allow_null=False, default=None)

    class Meta:
        fields = ["learner_id", "course_id", "activity_id", "session_id", "data", "instruments", "metadata",
                  "close_at", "close_session"]



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Model module. """
from .activity import Activity
from .activity_instrument import ActivityInstrument
from .alert import Alert
from .assessment_session import AssessmentSession
from .assessment_session_data import AssessmentSessionData
from .course import Course
from .course_group import CourseGroup
from .enrolment import Enrolment
from .enrolment_sample import EnrolmentSample
from .enrolment_sample_validation import EnrolmentSampleValidation
from .histogram_activity_instrument import HistogramActivityInstrument
from .histogram_activity_provider import HistogramActivityProvider
from .histogram_learner_instrument import HistogramLearnerInstrument
from .histogram_learner_provider import HistogramLearnerProvider
from .informed_consent import InformedConsent
from .informed_consent_document import InformedConsentDocument
from .institution import Institution
from .instructor import Instructor
from .instrument import Instrument
from .launcher import Launcher
from .learner import Learner
from .message import Message
from .message import MessageLocale
from .monitor import Monitor
from .provider import Provider
from .provider_notification import ProviderNotification
from .report_activity import ReportActivity
from .report_activity_session import ReportActivitySession
from .report_activity_instrument import ReportActivityInstrument
from .request import Request
from .request_provider_result import RequestProviderResult
from .request_result import RequestResult
from .send_category import SENDCategory
from .send_learner import SENDLearner
from .ui_option import UIOption
from .user import AuthenticatedModule
from .user import AuthenticatedUser
from .user import InstitutionUser
from .user import UnauthenticatedUser
from .user import User
from .vle import VLE
from .webhooks import WebhookClient
from .webhooks import WebhookMessage

__all__ = [
    'Activity',
    'ActivityInstrument',
    'Alert',
    'Course',
    'Enrolment',
    'EnrolmentSample',
    'EnrolmentSampleValidation',
    'InformedConsent',
    'InformedConsentDocument',
    'Instrument',
    'Learner',
    'Message',
    'MessageLocale',
    'Monitor',
    'Instructor',
    'Provider',
    'ProviderNotification',
    'Request',
    'RequestResult',
    'RequestProviderResult',
    'SENDCategory',
    'SENDLearner',
    'VLE',
    'Institution',
    'AssessmentSession',
    'AssessmentSessionData',
    'InstitutionUser',
    'UnauthenticatedUser',
    'AuthenticatedModule',
    'AuthenticatedUser',
    'Launcher',
    'UIOption',
    'User',
    'ReportActivity',
    'ReportActivitySession',
    'ReportActivityInstrument',
    'HistogramActivityInstrument',
    'HistogramActivityProvider',
    'HistogramLearnerProvider',
    'HistogramLearnerInstrument',
    'WebhookClient',
    'WebhookMessage',
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner API Alerts tasks module """
import simplejson
from celery.exceptions import Reject
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage

from tesla_ce import celery_app
from tesla_ce import models
from tesla_ce.models.alert import get_alert_from_value


@celery_app.task(ignore_result=True)
def create_alert(level, activity_id, learner_id, path, raised_at, instruments=None, session_id=None):
    try:
        # Store the alert message
        learner = models.Learner.objects.get(learner_id=learner_id)
        req = models.Alert.objects.create(level=get_alert_from_value(level),
                                          learner_id=learner.id,
                                          activity_id=activity_id,
                                          data=path,
                                          session_id=session_id,
                                          raised_at=raised_at)
        # Add the instruments
        if instruments is not None:
            req.instruments.set(instruments)
            if len(instruments) != req.instruments.count():
                error = {
                    "message": 'Invalid instruments',
                }
                default_storage.save('{}.error'.format(path), ContentFile(simplejson.dumps(error).encode('utf-8')))
                req.status = 2
                req.error_message = error["message"]
                req.save()
                raise Reject("Invalid Instruments", requeue=False)
        default_storage.save('{}.valid'.format(path), ContentFile(simplejson.dumps({
            "message": "stored",
            "alert_id": req.id
        }).encode('utf-8')))
        return req.id
    except models.Learner.DoesNotExist:
        raise Reject("Invalid Learner", requeue=False)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Enrolment sample request model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .instrument import Instrument
from .learner import Learner
from .message import Message

SAMPLE_STATUS = (
    (0, _('Stored')),
    (1, _('Valid')),
    (2, _('Error')),
    (3, _('Timeout')),
    (4, _('Missing Validator')),
)


class EnrolmentSample(BaseModel):
    """ Enrolment Sample Request model. """
    learner = models.ForeignKey(Learner, null=False,
                                on_delete=models.CASCADE,
                                help_text=_("Related learner"))

    status = models.IntegerField(choices=SAMPLE_STATUS, null=False,
                                 default=0,
                                 help_text=_('Status for this sample'))

    instruments = models.ManyToManyField(Instrument, help_text=_("Instruments this sample is collected for"))

    error_message = models.TextField(null=True, blank=None, default=None,
                                     help_text=_("Error message when status is error"))

    message_code = models.ForeignKey(Message, null=True, default=None,
                                     on_delete=models.SET_NULL,
                                     help_text=_("Related message code"))

    data = models.FileField(max_length=250, null=False, blank=False, help_text=_("Data path on storage."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<EnrolmentSample(learner='%s', status='%r')>" % (self.learner, self.status, )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Course views module """
import botocore
import json

from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.core.exceptions import ValidationError
from django.utils import timezone
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.views import Response
from rest_framework.views import status
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce import tasks

from tesla_ce.lib.exception import tesla_report_exception

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import VLECourseActivityInstrumentSerializer
from tesla_ce.apps.api.v2.serializers import VLECourseActivitySerializer
from tesla_ce.apps.api.v2.serializers import VLECourseActivityAttachmentSerializer
from tesla_ce.models import Activity


# pylint: disable=too-many-ancestors
class VLECourseActivityViewSet(NestedViewSetMixin, viewsets.ModelViewSet):
    """
    API endpoint that allows activity in a course to be viewed or edited.
    """
    model = Activity
    serializer_class = VLECourseActivitySerializer
    permission_classes = [
        permissions.VLEPermission
    ]
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_fields = ['vle_id', 'vle_activity_type', 'vle_activity_id', 'course_id', 'name']
    search_fields = ['vle_id', 'vle_activity_type', 'vle_activity_id', 'course_id', 'name']
    queryset = Activity.objects

    @action(detail=True, methods=['GET', 'POST',], url_path=r'attachment/(?P<learner_id>[\w|-]+)',
            serializer_class=VLECourseActivityAttachmentSerializer)
    def attachment(self, request, **kwargs):
        """
            Return the list of instruments (if any) waiting to process the attachment of the activity
        """
        activity = self.get_object()
        try:
            learner = activity.course.learners.get(learner_id=kwargs['learner_id'])
        except activity.course.learners.model.DoesNotExist:
            return Response("Learner does not exist", status=404)
        except ValidationError as verr:
            return Response(verr.__str__(), status=400)

        # For GET, just return the list of instruments waiting for the attachment
        if request.method == 'GET':
            activity_instruments = activity.get_learner_instruments(learner)

            # Filter instruments that accept attachments
            attachment_instruments = []
            for inst in activity_instruments:
                if inst.instrument_id in [4, 5]:
                    # Instruments that work with attachments (plag, fa)
                    attachment_instruments.append(inst)
                elif inst.instrument_id in [1, 3]:
                    # Instruments that can work with attachments (fr, vr)
                    options = inst.get_options()
                    if options is not None and 'offline' in options and options['offline']:
                        attachment_instruments.append(inst)
            return Response(VLECourseActivityInstrumentSerializer(instance=attachment_instruments, many=True).data)

        # POST action will store the attachment as a new request
        serializer = self.serializer_class(data=request.data, context={'request': request, 'view': self})
        try:
            if serializer.is_valid():
                # Get the filename
                metadata = serializer.data.get('metadata')
                if metadata is None:
                    filename = 'request'
                else:
                    filename = metadata.get('filename', 'document')
                # Generate the target path for the request
                path = '{}/requests/{}/{}/{}/documents/{}_{}'.format(
                    learner.institution_id,
                    learner.learner_id,
                    activity.course_id,
                    activity.id,
                    timezone.now().strftime("%Y%m%d_%H%M%S_%f"),
                    filename)

                # Add url parameters to data
                attachment_data = serializer.data
                attachment_data['learner_id'] = str(learner.learner_id)
                attachment_data['course_id'] = activity.course_id
                attachment_data['activity_id'] = activity.id

                # Remove extra data arguments
                if 'close_session' in attachment_data:
                    del attachment_data['close_session']
                if 'close_at' in attachment_data:
                    del attachment_data['close_at']

                # Store the request
                resp = default_storage.save(path, ContentFile(json.dumps(attachment_data).encode('utf-8')))

                # Notify the new stored request
                kwargs = {
                    'activity_id': activity.id,
                    'learner_id': str(learner.learner_id),
                    'path': path,
                    'instruments': serializer.data['instruments'],
                    'session_id': serializer.data.get('session_id')
                }
                tasks.requests.verification.create_request.apply_async(kwargs=kwargs)

                return Response({"status": "OK", "path": resp})

            return Response(status=status.HTTP_400_BAD_REQUEST,
                            data={"status": "ERROR", 'errors': serializer.errors},
                            content_type="application/json")
        except botocore.exceptions.ClientError as cli_err:
            tesla_report_exception(cli_err)
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            data={"status": "ERROR", 'errors': cli_err.response},
                            content_type="application/json")

        except Exception as ex:
            tesla_report_exception(ex)
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            data={"status": "ERROR", 'errors': ex},
                            content_type="application/json")
